import gc
import numpy as np
from anndata import AnnData
import anndata
import scipy.sparse as sp
import time
from sklearn.preprocessing import Normalizer, StandardScaler
import pickle
import pandas as pd
from . import utilities as ut
import sklearn.manifold as man
import sklearn.utils.sparsefuncs as sf
from packaging import version
import warnings
from numba.core.errors import NumbaWarning
warnings.filterwarnings("ignore", category=NumbaWarning)

__version__ = "0.8.7"

"""
Copyright 2018, Alexander J. Tarashansky, All rights reserved.
Email: <tarashan@stanford.edu>
"""


class SAM(object):
    """Self-Assembling Manifolds single-cell RNA sequencing analysis tool.

    SAM iteratively rescales the input gene expression matrix to emphasize
    genes that are spatially variable along the intrinsic manifold of the data.
    It outputs the gene weights, nearest neighbor matrix, and a 2D projection.

    Parameters
    ----------
    counts : tuple or list (scipy.sparse matrix, numpy.ndarray,numpy.ndarray),
        OR tuple or list (numpy.ndarray, numpy.ndarray,numpy.ndarray), OR
        pandas.DataFrame, OR anndata.AnnData

        If a tuple or list, it should contain the gene expression data
        (scipy.sparse or numpy.ndarray) matrix (cells x genes), numpy array of
        gene IDs, and numpy array of cell IDs in that order.

        If a pandas.DataFrame, it should be (cells x genes)

        Only use this argument if you want to pass in preloaded data. Otherwise
        use one of the load functions.


    Attributes
    ----------

    preprocess_args: dict
        Dictionary of arguments used for the 'preprocess_data' function.

    run_args: dict
        Dictionary of arguments used for the 'run' function.

    adata_raw: AnnData
        An AnnData object containing the raw, unfiltered input data.

    adata: AnnData
        An AnnData object containing all processed data and SAM outputs.

    """

    def __init__(self, counts=None, inplace=False):

        if isinstance(counts, tuple) or isinstance(counts, list):
            raw_data, all_gene_names, all_cell_names = counts
            if isinstance(raw_data, np.ndarray):
                raw_data = sp.csr_matrix(raw_data)

            self.adata_raw = AnnData(
                X=raw_data,
                obs={"obs_names": all_cell_names},
                var={"var_names": all_gene_names},
            )

        elif isinstance(counts, pd.DataFrame):
            raw_data = sp.csr_matrix(counts.values)
            all_gene_names = np.array(list(counts.columns.values))
            all_cell_names = np.array(list(counts.index.values))

            self.adata_raw = AnnData(
                X=raw_data,
                obs={"obs_names": all_cell_names},
                var={"var_names": all_gene_names},
            )

        elif isinstance(counts, AnnData):
            all_cell_names = np.array(list(counts.obs_names))
            all_gene_names = np.array(list(counts.var_names))
            if counts.is_view:
                counts = counts.copy()
                
            if inplace:
                self.adata_raw = counts
            else:
                self.adata_raw = counts.copy()            

        elif counts is not None:
            raise Exception(
                "'counts' must be either a tuple/list of "
                "(data,gene IDs,cell IDs) or a Pandas DataFrame of"
                "cells x genes"
            )

        if counts is not None:
            if np.unique(all_gene_names).size != all_gene_names.size:
                self.adata_raw.var_names_make_unique()
            if np.unique(all_cell_names).size != all_cell_names.size:
                self.adata_raw.obs_names_make_unique()

            if inplace:
                self.adata = self.adata_raw
            else:
                self.adata = self.adata_raw.copy()

            if "X_disp" not in self.adata_raw.layers.keys():
                self.adata.layers["X_disp"] = self.adata.X

        self.run_args = {}
        self.preprocess_args = {}

    def preprocess_data(
        self,
        div=1,
        downsample=0,
        sum_norm="cell_median",
        norm="log",
        min_expression=1,
        thresh_low=0.0,
        thresh_high=0.96,
        thresh = None,
        filter_genes=True,
    ):
        """Log-normalizes and filters the expression data.

        Parameters
        ----------

        div : float, optional, default 1
            The factor by which the gene expression will be divided prior to
            normalization (e.g. log normalization).

        downsample : float, optional, default 0
            The factor by which to randomly downsample the data. If 0, the
            data will not be downsampled.

        sum_norm : str or float, optional, default "cell_median"
            If a float, the total number of transcripts in each cell will be
            normalized to this value prior to normalization and filtering.
            Otherwise, nothing happens. If 'cell_median', each cell is
            normalized to have the median total read count per cell. If
            'gene_median', each gene is normalized to have the median total
            read count per gene.

        norm : str, optional, default 'log'
            If 'log', log-normalizes the expression data. If 'ftt', applies the
            Freeman-Tukey variance-stabilization transformation. If
            'multinomial', applies the Pearson-residual transformation (this is
            experimental and should only be used for raw, un-normalized UMI
            datasets). If None, the data is not normalized.

        min_expression : float, optional, default 1
            The threshold above which a gene is considered
            expressed. Gene expression values less than 'min_expression' are
            set to zero.

        thresh_low : float, optional, default 0.0
            Keep genes expressed in greater than 'thresh_low'*100 % of cells,
            where a gene is considered expressed if its expression value
            exceeds 'min_expression'.

        thresh_high : float, optional, default 0.96
            Keep genes expressed in less than 'thresh_high'*100 % of cells,
            where a gene is considered expressed if its expression value
            exceeds 'min_expression'.

        filter_genes : bool, optional, default True
            Setting this to False turns off filtering operations.

        """
        if thresh is not None:
            thresh_low = thresh
            thresh_high = 1-thresh

        self.preprocess_args = {
            "div": div,
            "sum_norm": sum_norm,
            "norm": norm,
            "min_expression": min_expression,
            "thresh_low": thresh_low,
            "thresh_high":thresh_high,
            "filter_genes": filter_genes,
        }
        
        self.run_args = self.adata.uns.get('run_args',{})
        # load data
        try:
            D = self.adata_raw.X
            self.adata = self.adata_raw.copy()

        except AttributeError:
            print("No data loaded")

        D = self.adata.X
        if isinstance(D, np.ndarray):
            D = sp.csr_matrix(D, dtype="float32")
        else:
            if str(D.dtype) != "float32":
                D = D.astype("float32")
            D.sort_indices()

        if D.getformat() == "csc":
            D = D.tocsr()

        # sum-normalize
        if sum_norm == "cell_median" and norm != "multinomial":
            s = D.sum(1).A.flatten()
            sum_norm = np.median(s)
            D = D.multiply(1 / s[:, None] * sum_norm).tocsr()
        elif sum_norm == "gene_median" and norm != "multinomial":
            s = D.sum(0).A.flatten()
            sum_norm = np.median(s[s > 0])
            s[s == 0] = 1
            D = D.multiply(1 / s[None, :] * sum_norm).tocsr()

        elif sum_norm is not None and norm != "multinomial":
            D = D.multiply(1 / D.sum(1).A.flatten()[:, None] * sum_norm).tocsr()

        # normalize
        self.adata.X = D
        if norm is None:
            D.data[:] = D.data / div

        elif norm.lower() == "log":
            D.data[:] = np.log2(D.data / div + 1)

        elif norm.lower() == "ftt":
            D.data[:] = np.sqrt(D.data / div) + np.sqrt(D.data / div + 1) - 1

        elif norm.lower() == "asin":
            D.data[:] = np.arcsinh(D.data / div)
        elif norm.lower() == "multinomial":
            ni = D.sum(1).A.flatten()  # cells
            pj = (D.sum(0) / D.sum()).A.flatten()  # genes
            col = D.indices
            row = []
            for i in range(D.shape[0]):
                row.append(i * np.ones(D.indptr[i + 1] - D.indptr[i]))
            row = np.concatenate(row).astype("int32")
            mu = sp.coo_matrix((ni[row] * pj[col], (row, col))).tocsr()
            mu2 = mu.copy()
            mu2.data[:] = mu2.data ** 2
            mu2 = mu2.multiply(1 / ni[:, None])
            mu.data[:] = (D.data - mu.data) / np.sqrt(mu.data - mu2.data)

            self.adata.X = mu
            if sum_norm is None:
                sum_norm = np.median(ni)
            D = D.multiply(1 / ni[:, None] * sum_norm).tocsr()
            D.data[:] = np.log2(D.data / div + 1)

        else:
            D.data[:] = D.data / div

        # zero-out low-expressed genes
        idx = np.where(D.data <= min_expression)[0]
        D.data[idx] = 0

        # filter genes
        idx_genes = np.arange(D.shape[1])
        if filter_genes:
            a, ct = np.unique(D.indices, return_counts=True)
            c = np.zeros(D.shape[1])
            c[a] = ct

            keep = np.where(
                np.logical_and(c / D.shape[0] > thresh_low, c / D.shape[0] <= thresh_high)
            )[0]

            idx_genes = np.array(list(set(keep) & set(idx_genes)))

        mask_genes = np.zeros(D.shape[1], dtype="bool")
        mask_genes[idx_genes] = True

        self.adata.X = self.adata.X.multiply(mask_genes[None, :]).tocsr()
        self.adata.X.eliminate_zeros()
        self.adata.var["mask_genes"] = mask_genes

        if norm == "multinomial":
            self.adata.layers["X_disp"] = D.multiply(mask_genes[None, :]).tocsr()
            self.adata.layers["X_disp"].eliminate_zeros()
        else:
            self.adata.layers["X_disp"] = self.adata.X

        self.calculate_mean_var()

        self.adata.uns["preprocess_args"] = self.preprocess_args
        self.adata.uns['run_args'] = self.run_args

    def calculate_mean_var(self):
        mu,var = sf.mean_variance_axis(self.adata.X,axis=0)
        self.adata.var['means'] = mu
        self.adata.var['variances'] = var
        
    def get_avg_obsm(self, keym, keyl):
        clu = self.get_labels_un(keyl)
        cl = self.get_labels(keyl)
        x = []
        for i in range(clu.size):
            x.append(self.adata.obsm[keym][cl == clu[i]].mean(0))
        x = np.vstack(x)
        return x

    def get_labels_un(self, key):
        if key not in list(self.adata.obs.keys()):
            print("Key does not exist in `obs`.")
            return np.array([])
        else:
            return np.array(list(np.unique(self.adata.obs[key])))

    def get_labels(self, key):
        if key not in list(self.adata.obs.keys()):
            print("Key does not exist in `obs`.")
            return np.array([])
        else:
            return np.array(list(self.adata.obs[key]))

    def get_cells(self, label, key):
        """Retrieves cells of a particular annotation.

        Parameters
        ----------
        label - The annotation to retrieve
        key - The key in `obs` from which to retrieve the annotation.

        """
        if key not in list(self.adata.obs.keys()):
            print("Key does not exist in `obs`.")
            return np.array([])
        else:
            return np.array(
                list(self.adata.obs_names[np.array(list(self.adata.obs[key])) == label])
            )

    def load_data(
        self,
        filename,
        transpose=True,
        save_sparse_file=None,
        sep=",",
        calculate_avg=False,
        **kwargs
    ):
        """Loads the specified data file. The file can be a table of
        read counts (i.e. '.csv' or '.txt'), with genes as rows and cells
        as columns by default. The file can also be a pickle file (output from
        'save_sparse_data') or an h5ad file (output from 'save_anndata').

        This function that loads the file specified by 'filename'.

        Parameters
        ----------
        filename - string
            The path to the tabular raw expression counts file.

        sep - string, optional, default ','
            The delimeter used to read the input data table. By default
            assumes the input table is delimited by commas.

        save_sparse_file - str, optional, default None
            Path to which the sparse data will be saved ('.h5ad').
            Writes the SAM 'adata_raw' object to a h5ad file (the native AnnData
            file format) for faster loading in the future.

        transpose - bool, optional, default True
            By default, assumes file is (genes x cells). Set this to False if
            the file has dimensions (cells x genes).

        calculate_avg - bool, optional, default True
            If nearest neighbors are already calculated in the .h5ad file,
            setting this parameter to True performs knn averaging and stores
            the result in adata.layers['X_knn_avg']. This is a fairly dense
            matrix, so set this to False if you do not need the averaged
            expressions.

        """
        if filename.split(".")[-1] == "p":
            raw_data, all_cell_names, all_gene_names = pickle.load(open(filename, "rb"))

            if transpose:
                raw_data = raw_data.T
                if raw_data.getformat() == "csc":
                    print("Converting sparse matrix to csr format...")
                    raw_data = raw_data.tocsr()

            save_sparse_file = None
        elif filename.split(".")[-1] != "h5ad":
            df = pd.read_csv(filename, sep=sep, index_col=0, **kwargs)
            if transpose:
                dataset = df.T
            else:
                dataset = df

            raw_data = sp.csr_matrix(dataset.values)
            all_cell_names = np.array(list(dataset.index.values))
            all_gene_names = np.array(list(dataset.columns.values))

        if filename.split(".")[-1] != "h5ad":
            self.adata_raw = AnnData(
                X=raw_data,
                obs={"obs_names": all_cell_names},
                var={"var_names": all_gene_names},
            )

            if np.unique(all_gene_names).size != all_gene_names.size:
                self.adata_raw.var_names_make_unique()
            if np.unique(all_cell_names).size != all_cell_names.size:
                self.adata_raw.obs_names_make_unique()

            self.adata = self.adata_raw.copy()
            self.adata.layers["X_disp"] = raw_data

        else:
            self.adata = anndata.read_h5ad(filename, **kwargs)
            if self.adata.raw is not None:
                self.adata_raw = AnnData(X=self.adata.raw.X)
                self.adata_raw.var_names = self.adata.raw.var_names
                self.adata_raw.obs_names = self.adata.obs_names
                self.adata_raw.obs = self.adata.obs

                if version.parse(str(anndata.__version__)) >= version.parse("0.7rc1"):
                    del self.adata.raw
                else:
                    self.adata.raw = None

                if (
                    "X_knn_avg" not in self.adata.layers.keys()
                    and "connectivities" in self.adata.obsp.keys()
                    and calculate_avg
                ):
                    self.dispersion_ranking_NN(save_avgs=True)
            else:
                self.adata_raw = self.adata

            if "X_disp" not in list(self.adata.layers.keys()):
                self.adata.layers["X_disp"] = self.adata.X
            save_sparse_file = None
        filename = '.'.join(filename.split('.')[:-1])+'.h5ad'
        self.adata.uns['path_to_file'] = filename
        self.adata_raw.uns['path_to_file'] = filename        
        if save_sparse_file is not None:
            if save_sparse_file.split(".")[-1] == "p":
                self.save_sparse_data(save_sparse_file)
            elif save_sparse_file.split(".")[-1] == "h5ad":
                self.save_anndata(save_sparse_file)

    def save_anndata(self, fname='', save_knn=False, **kwargs):
        """Saves `adata_raw` to a .h5ad file (AnnData's native file format).

        Parameters
        ----------
        fname - string, default ''
            The filename of the output file. If empty, defaults to `.adata.uns['path_to_file']`.

        save_knn - bool, optional, default = False
            If True, saves `.layers['X_knn_avg']`. If False, does not save
            this layer. Default value is set to False as the nearest-neighbor
            averaged expression values can be quite dense.

        """
        if not save_knn:
            try:
                Xknn = self.adata.layers["X_knn_avg"]
                del self.adata.layers["X_knn_avg"]
            except:
                0
        if fname == '':
            try:
                fname = self.adata.uns['path_to_file']
            except KeyError:
                raise KeyError('Path to file not known.')
                
        x = self.adata
        x.raw = self.adata_raw
        
        # fixing weird issues when index name is an integer.
        for y in [x.obs.columns,x.var.columns,
                  x.obs.index,x.var.index,
                  x.raw.var.index,x.raw.var.columns]:
            y.name = str(y.name) if y.name is not None else None

        x.write_h5ad(fname, **kwargs)
        if version.parse(str(anndata.__version__)) >= version.parse("0.7rc1"):
            del x.raw
        else:
            x.raw = None

        try:
            self.adata.layers["X_knn_avg"] = Xknn
        except:
            0

    def load_var_annotations(self, aname, sep=",", key_added="annotations"):
        """Loads gene annotations.

        Loads the gene annotations into .adata_raw.var and .adata.var. The keys
        added correspond to the column labels in the input table.

        Parameters
        ----------
        aname - string or pandas.DataFrame
            If string, it is the path to the annotations file, which should be
            a table with the first column being the gene IDs and the first row
            being the column names for the annotations. Alternatively, you can
            directly pass in a preloaded pandas.DataFrame.

        sep - string, default ','
            The delimeter used in the file. Ignored if passing in a preloaded
            pandas.DataFrame.

        """
        if isinstance(aname, pd.DataFrame):
            ann = aname
        else:
            ann = pd.read_csv(aname, sep=sep, index_col=0)

        for i in range(ann.shape[1]):
            self.adata_raw.var[ann.columns[i]] = ann[ann.columns[i]]
            self.adata.var[ann.columns[i]] = ann[ann.columns[i]]

    def load_obs_annotations(self, aname, sep=","):
        """Loads cell annotations.

        Loads the cell annotations into .adata_raw.obs and .adata.obs. The keys
        added correspond to the column labels in the input table.

        Parameters
        ----------
        aname - string or pandas.DataFrame
            If string, it is the path to the annotations file, which should be
            a table with the first column being the cell IDs and the first row
            being the column names for the annotations. Alternatively, you can
            directly pass in a preloaded pandas.DataFrame.

        sep - string, default ','
            The delimeter used in the file. Ignored if passing in a preloaded
            pandas.DataFrame.

        """
        if isinstance(aname, pd.DataFrame):
            ann = aname
        else:
            ann = pd.read_csv(aname, sep=sep, index_col=0)

        for i in range(ann.shape[1]):
            self.adata_raw.obs[ann.columns[i]] = ann[ann.columns[i]]
            self.adata.obs[ann.columns[i]] = ann[ann.columns[i]]

    def scatter(
        self,
        projection=None,
        c=None,
        colorspec=None,
        cmap="rainbow",
        linewidth=0.0,
        edgecolor="k",
        axes=None,
        colorbar=True,
        s=10,
        **kwargs
    ):

        """Display a scatter plot.
        Displays a scatter plot using the SAM projection or another input
        projection.

        Parameters
        ----------
        projection - string, numpy.ndarray, default None
            A case-sensitive string indicating the projection to display (a key
            in adata.obsm) or a 2D numpy array with cell coordinates. If None,
            projection defaults to UMAP.

        c - string, numpy.ndarray, default None
            Categorical data to be mapped to a colormap and overlaid on top of
            the projection. Can be a key from adata.obs or a 1D numpy array.

        colorspec - string, numpy.ndarray, default None
            A string specifying a color or an array specifying the color
            for each point (can be strings, RGBA, RGB, etc). Colorbar will be
            turned off if `colorspec` is not None.

        axes - matplotlib axis, optional, default None
            Plot output to the specified, existing axes. If None, create new
            figure window.

        **kwargs - all keyword arguments in matplotlib.pyplot.scatter are eligible.
        """

        try:
            import matplotlib.pyplot as plt

            if isinstance(projection, str):
                try:
                    dt = self.adata.obsm[projection]
                except KeyError:
                    print(
                        "Please create a projection first using run_umap or" "run_tsne"
                    )

            elif projection is None:
                try:
                    dt = self.adata.obsm["X_umap"]
                except KeyError:
                    try:
                        dt = self.adata.obsm["X_tsne"]
                    except KeyError:
                        print(
                            "Please create either a t-SNE or UMAP projection" "first."
                        )
                        return
            else:
                dt = projection

            if axes is None:
                plt.figure()
                axes = plt.gca()

            if colorspec is not None:
                axes.scatter(
                    dt[:, 0],
                    dt[:, 1],
                    s=s,
                    linewidth=linewidth,
                    edgecolor=edgecolor,
                    c=colorspec,
                    **kwargs
                )
            elif c is None:
                axes.scatter(
                    dt[:, 0],
                    dt[:, 1],
                    s=s,
                    linewidth=linewidth,
                    edgecolor=edgecolor,
                    **kwargs
                )
            else:

                if isinstance(c, str):
                    try:
                        c = self.get_labels(c)
                    except KeyError:
                        0  # do nothing

                if (isinstance(c[0], str) or isinstance(c[0], np.str_)) and (
                    isinstance(c, np.ndarray) or isinstance(c, list)
                ):
                    i = ut.convert_annotations(c)
                    ui, ai = np.unique(i, return_index=True)
                    cax = axes.scatter(
                        dt[:, 0],
                        dt[:, 1],
                        c=i,
                        cmap=cmap,
                        s=s,
                        linewidth=linewidth,
                        edgecolor=edgecolor,
                        **kwargs
                    )

                    if colorbar:
                        cbar = plt.colorbar(cax, ax=axes, ticks=ui)
                        cbar.ax.set_yticklabels(c[ai])
                else:
                    if not (isinstance(c, np.ndarray) or isinstance(c, list)):
                        colorbar = False
                    i = c

                    cax = axes.scatter(
                        dt[:, 0],
                        dt[:, 1],
                        c=i,
                        cmap=cmap,
                        s=s,
                        linewidth=linewidth,
                        edgecolor=edgecolor,
                        **kwargs
                    )

                    if colorbar:
                        plt.colorbar(cax, ax=axes)
            return axes
        except ImportError:
            print("matplotlib not installed!")

    def show_gene_expression(self, gene, avg=True, axes=None, **kwargs):
        """Display a gene's expressions.
        Displays a scatter plot using the SAM projection or another input
        projection with a particular gene's expressions overlaid.
        Parameters
        ----------
        gene - string
            a case-sensitive string indicating the gene expression pattern
            to display.
        avg - bool, optional, default True
            If True, the plots use the k-nearest-neighbor-averaged expression
            values to smooth out noisy expression patterns and improves
            visualization.
        axes - matplotlib axis, optional, default None
            Plot output to the specified, existing axes. If None, create new
            figure window.
        **kwargs - all keyword arguments in 'SAM.scatter' are eligible.
        """
        all_gene_names = np.array(list(self.adata.var_names))
        cell_names = np.array(list(self.adata.obs_names))
        all_cell_names = np.array(list(self.adata_raw.obs_names))
        idx2 = np.where(np.in1d(all_cell_names, cell_names))[0]
        idx = np.where(all_gene_names == gene)[0]
        name = gene
        if idx.size == 0:
            print(
                "Gene note found in the filtered dataset. Note that genes "
                "are case sensitive."
            )
            return

        if avg:
            a = self.adata.layers["X_knn_avg"][:, idx].toarray().flatten()
            if a.sum() == 0:
                a = self.adata_raw.X[:, idx].toarray().flatten()[idx2]
                try:
                    norm = self.preprocess_args["norm"]
                except KeyError:
                    norm = "log"
                if norm is not None:
                    if norm.lower() == "log":
                        a = np.log2(a + 1)

                    elif norm.lower() == "ftt":
                        a = np.sqrt(a) + np.sqrt(a + 1)
                    elif norm.lower() == "asin":
                        a = np.arcsinh(a)
        else:
            a = self.adata_raw.X[:, idx].toarray().flatten()[idx2]
            try:
                norm = self.preprocess_args["norm"]
            except KeyError:
                norm = "log"

            if norm is not None:
                if norm.lower() == "log":
                    a = np.log2(a + 1)

                elif norm.lower() == "ftt":
                    a = np.sqrt(a) + np.sqrt(a + 1)
                elif norm.lower() == "asin":
                    a = np.arcsinh(a)

        axes = self.scatter(c=a, axes=axes, **kwargs)
        axes.set_title(name)

        return axes, a

    def dispersion_ranking_NN(self, nnm=None, num_norm_avg=50, weight_mode='combined',save_avgs=False):
        """Computes the spatial dispersion factors for each gene.

        Parameters
        ----------
        nnm - scipy.sparse, default None
            Square cell-to-cell nearest-neighbor matrix. If None, uses the
            nearest neighbor matrix in .adata.obsp['connectivities']

        num_norm_avg - int, optional, default 50
            The top 'num_norm_avg' dispersions are averaged to determine the
            normalization factor when calculating the weights. This ensures
            that outlier genes do not significantly skew the weight
            distribution.

        Returns:
        -------
        weights - ndarray, float
            The vector of gene weights.
        """
        if nnm is None:
            nnm = self.adata.obsp["connectivities"]
        f = nnm.sum(1).A
        f[f==0]=1
        D_avg = (nnm.multiply(1 / f)).dot(self.adata.layers["X_disp"])

        if save_avgs:
            self.adata.layers["X_knn_avg"] = D_avg.copy()

        if sp.issparse(D_avg):
            mu, var = sf.mean_variance_axis(D_avg, axis=0)            
            if weight_mode == 'rms':
                D_avg.data[:]=D_avg.data**2
                mu,_ =sf.mean_variance_axis(D_avg, axis=0)
                mu=mu**0.5
                
            if weight_mode == 'combined':
                D_avg.data[:]=D_avg.data**2
                mu2,_ =sf.mean_variance_axis(D_avg, axis=0)
                mu2=mu2**0.5                
        else:
            mu = D_avg.mean(0)
            var = D_avg.var(0)
            if weight_mode == 'rms':
                mu =(D_avg**2).mean(0)**0.5
            if weight_mode == 'combined':
                mu2 =(D_avg**2).mean(0)**0.5
            
                
        if not save_avgs:        
            del D_avg
            gc.collect()

        if weight_mode == 'dispersion' or weight_mode == 'rms' or weight_mode == 'combined':
            dispersions = np.zeros(var.size)
            dispersions[mu > 0] = var[mu > 0] / mu[mu > 0]
            self.adata.var["spatial_dispersions"] = dispersions.copy()
            
            if weight_mode == 'combined':
                dispersions2 = np.zeros(var.size)
                dispersions2[mu2 > 0] = var[mu2 > 0] / mu2[mu2 > 0]
                

        elif weight_mode == 'variance':
            dispersions = var
            self.adata.var["spatial_variances"] = dispersions.copy()
        else:
            raise ValueError('`weight_mode` ',weight_mode,' not recognized.')

        ma = np.sort(dispersions)[-num_norm_avg:].mean()
        dispersions[dispersions >= ma] = ma

        weights = ((dispersions / dispersions.max()) ** 0.5).flatten()
        
        if weight_mode == 'combined':
            ma = np.sort(dispersions2)[-num_norm_avg:].mean()
            dispersions2[dispersions2 >= ma] = ma

            weights2 = ((dispersions2 / dispersions2.max()) ** 0.5).flatten()
            weights = np.vstack((weights,weights2)).max(0)

        return weights

    def run(
        self,
        max_iter=10,
        verbose=True,
        projection="umap",
        stopping_condition=1e-2,
        num_norm_avg=50,
        k=20,
        distance="cosine",
        preprocessing="StandardScaler",
        npcs=150,
        n_genes=3000,
        weight_PCs=False,
        sparse_pca=False,
        proj_kwargs={},
        seed = 0,
        weight_mode='rms',
        components=None
    ):
        """Runs the Self-Assembling Manifold algorithm.

        Parameters
        ----------
        k - int, optional, default 20
            The number of nearest neighbors to identify for each cell.

        distance : string, optional, default 'correlation'
            The distance metric to use when identifying nearest neighbors.
            Can be any of the distance metrics supported by sklearn's 'pdist'.

        max_iter - int, optional, default 10
            The maximum number of iterations SAM will run.

        stopping_condition - float, optional, default 5e-3
            The stopping condition threshold for the RMSE between gene weights
            in adjacent iterations.

        verbose - bool, optional, default True
            If True, the iteration number and error between gene weights in
            adjacent iterations will be displayed.

        projection - str, optional, default 'umap'
            If 'tsne', generates a t-SNE embedding. If 'umap', generates a UMAP
            embedding. Otherwise, no embedding will be generated.

        preprocessing - str, optional, default 'StandardScaler'
            If 'Normalizer', use sklearn.preprocessing.Normalizer, which
            normalizes expression data prior to PCA such that each cell has
            unit L2 norm. If 'StandardScaler', use
            sklearn.preprocessing.StandardScaler, which normalizes expression
            data prior to PCA such that each gene has zero mean and unit
            variance. Otherwise, do not normalize the expression data. We
            recommend using 'StandardScaler' for large datasets and
            'Normalizer' otherwise.
        
        npcs - int, optional, default 150
            Sets the number of principal components to use.
            
        n_genes - int, optional, default 3000
            Sets the number of genes to include prior to PCA. Genes are selected
            based on their SAM weights. If `None`, all genes are selected.
            In this case, it is recommended to set `sparse_pca` equal to True.
            
        num_norm_avg - int, optional, default 50
            The top 'num_norm_avg' dispersions are averaged to determine the
            normalization factor when calculating the weights. This prevents
            genes with large spatial dispersions from skewing the distribution
            of weights.

        sparse_pca - bool, optional, default False
            If True, uses an implementation of PCA that accepts sparse inputs.
            This is worth setting True for large datasets, where memory
            constraints start becoming noticeable.

        weight_PCs - bool, optional, default False
            Scale the principal components by their eigenvalues. If many
            cell populations are expected, it is recommended to set this False
            as the populations may be found on lesser-varying PCs.
            
        weight_mode - str, optional, default 'combined'
            Can be one of 'dispersion', 'variance', 'rms', 'combined'.
                'dispersion' - Gene weights are calculated as the fano factors
                (variance / mean) of kNN-averaged expressions.
                
                'variance' - Gene weights are calculated as the variances of
                kNN-averaged expressions.
                
                'rms' - Gene weights are calculated as a modified fano factor
                (variance / root-mean square) of kNN-averaged expressions.
                This makes the weight more robust to genes with extremely low
                mean expressions.
                
                'combined' - Gene weights are calculated as the maximum of the 
                weights calculated from 'dispersion' and 'rms' modes.
                

        proj_kwargs - dict, optional, default {}
            A dictionary of keyword arguments to pass to the projection
            functions.
        """
        D = self.adata.X
        if k < 5:
            k = 5
        if k > D.shape[0] - 1:
            k = D.shape[0] - 2

        if preprocessing not in ["StandardScaler", "Normalizer", None, "None"]:
            raise ValueError(
                "preprocessing must be 'StandardScaler', 'Normalizer', or None"
            )
        if weight_mode not in ["dispersion", "variance", "rms", "combined"]:
            raise ValueError(
                "weight_mode must be 'dispersion', 'variance', 'rms', or 'combined'."
            )

        if self.adata.layers['X_disp'].min() < 0 and weight_mode == 'dispersion':
            print("`X_disp` layer contains negative values. Setting `weight_mode` to 'rms'.")
            weight_mode = 'rms'

        numcells = D.shape[0]

        if n_genes == None:
            n_genes = self.adata.shape[1]
            if not sparse_pca and numcells > 10000:
                warnings.warn('All genes are being used. It is recommended '
                              'to set `sparse_pca=False` to satisfy memory '
                              'constraints for datasets with more than '
                              '10,000 cells. Setting `sparse_pca` to True.')
                sparse_pca=True

        if not sparse_pca:
            n_genes = min(n_genes, (D.sum(0) > 0).sum())
            
        self.run_args = {
            "max_iter": max_iter,
            "verbose": verbose,
            "projection": projection,
            "stopping_condition": stopping_condition,
            "num_norm_avg": num_norm_avg,
            "k": k,
            "distance": distance,
            "preprocessing": preprocessing,
            "npcs": npcs,
            "n_genes": n_genes,
            "weight_PCs": weight_PCs,
            "proj_kwargs": proj_kwargs,
            "sparse_pca": sparse_pca,
            "weight_mode": weight_mode,
            "seed": seed,
            "components": components,
        }
        self.adata.uns["run_args"] = self.run_args
        
        
        tinit = time.time()
        np.random.seed(seed)

        if verbose:
            print("RUNNING SAM")

        W = np.ones(D.shape[1])
        self.adata.var['weights'] = W

        old = np.zeros(W.size)
        new = W

        i = 0
        err = ((new - old) ** 2).mean() ** 0.5

        if max_iter < 5:
            max_iter = 5

        nnas = num_norm_avg

        while i < max_iter and err > stopping_condition:

            conv = err
            if verbose:
                print("Iteration: " + str(i) + ", Convergence: " + str(conv))

            i += 1
            old = new
            if i == 1:                
                first=True
            else:
                first=False
            W = self.calculate_nnm(
                n_genes=n_genes, preprocessing=preprocessing, npcs=npcs, num_norm_avg=nnas,
                weight_PCs=weight_PCs, sparse_pca=sparse_pca,weight_mode=weight_mode,seed=seed,components=components,first=first
                )                
            gc.collect()
            new = W
            err = ((new - old) ** 2).mean() ** 0.5
            self.adata.var['weights']=W

        all_gene_names = np.array(list(self.adata.var_names))
        indices = np.argsort(-W)
        ranked_genes = all_gene_names[indices]

        self.adata.uns["ranked_genes"] = ranked_genes

        if projection == "tsne":
            if verbose:
                print("Computing the t-SNE embedding...")
            self.run_tsne(**proj_kwargs)
        elif projection == "umap":
            if verbose:
                print("Computing the UMAP embedding...")
            self.run_umap(seed=seed,**proj_kwargs)
        elif projection == "diff_umap":
            if verbose:
                print("Computing the diffusion UMAP embedding...")
            self.run_diff_umap(**proj_kwargs)

        elapsed = time.time() - tinit
        if verbose:
            print("Elapsed time: " + str(elapsed) + " seconds")

    def calculate_nnm(
        self, n_genes=3000, preprocessing='StandardScaler', npcs=150, num_norm_avg=50, weight_PCs=False, sparse_pca=False,
        update_manifold=True,weight_mode='dispersion',seed=0,components=None,first=False
    ):
        if 'means' not in self.adata.var.keys() or 'variances' not in self.adata.var.keys():
            print('Computing means and variances of genes.')
            self.calculate_mean_var()

        numcells = self.adata.shape[0]
        D = self.adata.X
        W = self.adata.var["weights"].values

        k = self.adata.uns['run_args'].get("k", 20)
        distance = self.adata.uns['run_args'].get("distance", "correlation")

        if n_genes is None:
            gkeep = np.arange(W.size)
        else:
            if first:
                mu = np.array(list(self.adata.var['means']))
                var = np.array(list(self.adata.var['variances']))
                mu[mu==0]=1
                dispersions = var/mu
                gkeep = np.sort(np.argsort(-dispersions)[:n_genes])
            else:    
                gkeep = np.sort(np.argsort(-W)[:n_genes])

        if preprocessing == "Normalizer":
            Ds = D[:, gkeep]
            if sp.issparse(Ds) and not sparse_pca:
                Ds = Ds.toarray()

            Ds = Normalizer().fit_transform(Ds)

        elif preprocessing == "StandardScaler":
            if not sparse_pca:
                Ds = D[:, gkeep]
                if sp.issparse(Ds):
                    Ds = Ds.toarray()

                v = self.adata.var['variances'].values[gkeep]
                m = self.adata.var['means'].values[gkeep]
                v[v==0]=1
                Ds = (Ds - m)/v**0.5

                Ds[Ds > 10] = 10
                Ds[Ds < -10] = -10
            else:
                Ds = D[:, gkeep]
                v = self.adata.var['variances'].values[gkeep]
                v[v==0]=1
                Ds = Ds.multiply(1/v**0.5).tocsr()

        else:
            Ds = D[:, gkeep].toarray()

        if sp.issparse(Ds):
            D_sub = Ds.multiply(W[gkeep]).tocsr()
        else:
            D_sub = Ds * (W[gkeep])

        if components is None:
            if not sparse_pca:
                npcs = min(npcs, min((D.shape[0],gkeep.size)))
                if numcells > 500:
                    g_weighted, pca = ut.weighted_PCA(
                        D_sub,
                        npcs=npcs,
                        do_weight=weight_PCs,
                        solver="auto",seed=seed
                    )
                else:
                    g_weighted, pca = ut.weighted_PCA(
                        D_sub,
                        npcs=npcs,
                        do_weight=weight_PCs,
                        solver="full",seed=seed
                    )
                self.pca_obj = pca
                self.components = pca.components_

            else:
                npcs=min(npcs, min((D.shape[0],gkeep.size)) - 1)
                v = self.adata.var['variances'].values[gkeep]
                v[v==0]=1
                m = self.adata.var['means'].values[gkeep]*W[gkeep]
                no = (m/v**0.5) if preprocessing=='StandardScaler' else D_sub.mean(0).A.flatten()
                output = ut._pca_with_sparse(D_sub, npcs,mu = (no)[None,:], seed = seed)
                self.components = output['components']
                g_weighted = output['X_pca']

                if weight_PCs:
                    ev = output['variance']
                    ev = ev / ev.max()
                    g_weighted = g_weighted * (ev ** 0.5)
        else:
            self.components = components[:,gkeep]
            v = self.adata.var['variances'].values[gkeep]
            v[v==0]=1
            m = self.adata.var['means'].values[gkeep]*W[gkeep]
            ns = (m/v**0.5) if preprocessing=='StandardScaler' else D_sub.mean(0).A.flatten()
            if sp.issparse(D_sub):
                g_weighted = D_sub.dot(self.components.T) - ns.flatten().dot(self.components.T)
            else:
                g_weighted = (D_sub-ns).dot(self.components.T)
            if weight_PCs:
                ev = g_weighted.var(0) #fix this?
                ev = ev / ev.max()
                g_weighted = g_weighted * (ev ** 0.5)

        if distance == "euclidean":
            pass;#g_weighted = Normalizer().fit_transform(g_weighted)

        if update_manifold:
            edm = ut.calc_nnm(g_weighted, k, distance)
            self.adata.obsp['distances'] = edm.tolil()
            self.adata.obsp['distances'].setdiag(0)
            self.adata.obsp['distances']=self.adata.obsp['distances'].tocsr()
            
            EDM = edm.copy()
            EDM.data[:] = 1
            EDM = EDM.tolil(); EDM.setdiag(1); EDM = EDM.tocsr();

            self.adata.obsp['connectivities'] = EDM

            if distance in ['correlation','cosine']: #keep edge weights and store in nnm if distance is bounded
                edm.data[:] = 1-edm.data
                edm = edm.tolil(); edm.setdiag(1); edm = edm.tocsr();
                edm.data[edm.data<0]=0.001 #if negative correlation, set close to zero but not zero to preserve kNN structure
                self.adata.obsp['nnm'] = edm
            else:
                self.adata.obsp['nnm'] = EDM
            W = self.dispersion_ranking_NN(EDM, weight_mode=weight_mode, num_norm_avg=num_norm_avg)
            self.adata.obsm["X_pca"] = g_weighted
            ge = np.array(list(self.adata.var_names[gkeep]))
            self.X_processed = (D_sub, ge, gkeep)
            self.adata.varm["PCs"] = np.zeros(shape=(self.adata.n_vars, npcs))
            self.adata.varm["PCs"][gkeep] = self.components.T
            return W
        else:
            print('Not updating the manifold...')
            PCs = np.zeros(shape=(self.adata.n_vars, npcs))
            PCs[gkeep] = self.components.T
            return PCs,g_weighted

    def run_tsne(self, X=None, metric="correlation", **kwargs):
        """Wrapper for sklearn's t-SNE implementation.

        See sklearn for the t-SNE documentation. All arguments are the same
        with the exception that 'metric' is set to 'correlation' by default.
        """
        if X is not None:
            dt = man.TSNE(metric=metric, **kwargs).fit_transform(X)
            return dt

        else:
            distance = self.adata.uns['run_args'].get("distance", "correlation")
            dt = man.TSNE(metric=distance, **kwargs).fit_transform(
                self.adata.obsm["X_pca"]
            )
            tsne2d = dt
            self.adata.obsm["X_tsne"] = tsne2d

    def run_umap(self, X="X_pca", metric=None,seed = None, **kwargs):
        """Wrapper for umap-learn.

        See https://github.com/lmcinnes/umap sklearn for the documentation
        and source code.
        """

        import umap as umap

        if metric is None:
            metric = self.adata.uns['run_args'].get("distance", "correlation")
            
        if seed is None:
            seed = self.adata.uns['run_args'].get('seed',0)

        if type(X) is str:
            if X == "":
                X = self.adata.X
            else:
                X = self.adata.obsm[X]
            # print(X.shape)
            umap_obj = umap.UMAP(metric=metric,random_state=seed, **kwargs)
            umap2d = umap_obj.fit_transform(X)
            self.adata.obsm["X_umap"] = umap2d
            return umap_obj
        else:
            umap_obj = umap.UMAP(metric=metric,random_state=seed, **kwargs)
            dt = umap_obj.fit_transform(X)
            return dt, umap_obj

    def run_diff_umap(
        self, use_rep="X_pca", metric="euclidean", n_comps=15, method="gauss", **kwargs
    ):
        """
        Experimental -- running UMAP on the diffusion components. Requires scanpy.
        """
        import scanpy.api as sc

        k = self.adata.uns['run_args'].get("k", 20)
        distance = self.adata.uns['run_args'].get("distance", "correlation")
        sc.pp.neighbors(
            self.adata, use_rep=use_rep, n_neighbors=k, metric=distance, method=method
        )
        sc.tl.diffmap(self.adata, n_comps=n_comps)
        sc.pp.neighbors(
            self.adata,
            use_rep="X_diffmap",
            n_neighbors=k,
            metric="euclidean",
            method=method,
        )

        if "X_umap" in self.adata.obsm.keys():
            temp = self.adata.obsm["X_umap"].copy()

        sc.tl.umap(self.adata, min_dist=0.1, copy=False)
        temp2 = self.adata.obsm["X_umap"]
        self.adata.obsm["X_umap"] = temp
        self.adata.obsm["X_diff_umap"] = temp2

    def run_diff_map(
        self, use_rep="X_pca", metric="euclidean", n_comps=15, method="gauss", **kwargs
    ):
        import scanpy.api as sc

        k = self.adata.uns['run_args'].get("k", 20)
        distance = self.adata.uns['run_args'].get("distance", "correlation")
        sc.pp.neighbors(
            self.adata, use_rep=use_rep, n_neighbors=k, metric=distance, method=method
        )
        sc.tl.diffmap(self.adata, n_comps=n_comps + 1)
        self.adata.obsm["X_diffmap"] = self.adata.obsm["X_diffmap"][:, 1:]

    def density_clustering(self, X=None, eps=1, metric="euclidean", **kwargs):
        from sklearn.cluster import DBSCAN

        if X is None:
            X = self.adata.obsm["X_umap"]
            save = True
        else:
            save = False

        cl = DBSCAN(eps=eps, metric=metric, **kwargs).fit_predict(X)
        k = self.adata.uns['run_args'].get("k", 20)
        idx0 = np.where(cl != -1)[0]
        idx1 = np.where(cl == -1)[0]
        if idx1.size > 0 and idx0.size > 0:
            xcmap = ut.generate_euclidean_map(X[idx0, :], X[idx1, :])
            knn = np.argsort(xcmap.T, axis=1)[:, :k]
            nnm = np.zeros(xcmap.shape).T
            nnm[
                np.tile(np.arange(knn.shape[0])[:, None], (1, knn.shape[1])).flatten(),
                knn.flatten(),
            ] = 1
            nnmc = np.zeros((nnm.shape[0], cl.max() + 1))
            for i in range(cl.max() + 1):
                nnmc[:, i] = nnm[:, cl[idx0] == i].sum(1)

            cl[idx1] = np.argmax(nnmc, axis=1)

        if save:
            self.adata.obs["dbscan_clusters"] = pd.Categorical(cl)
        else:
            return cl

    def clustering(self, X=None, param=None, method="leiden"):
        """A wrapper for clustering the SAM output using various clustering
        algorithms

        Parameters
        ----------
        X - data, optional, default None
            Data to be passed into the selected clustering algorithm. If None,
            uses the data in the SAM AnnData object. Different clustering
            algorithms accept different types of data. For example, Louvain
            and Leiden accept a scipy.sparse adjacency matrix representing the
            nearest neighbor graph. Dbscan, kmeans, and Hdbscan accept
            coordinates (like UMAP coordinates or PCA coordinates). If None,
            cluster results are saved to the 'obs' attribute in the AnnData
            object. Otherwise, cluster results are returned.

        param : float, optional, default None
            The parameter used for the different clustering algorithms. For
            louvain and leiden, it is the resolution parameter. For dbscan, it
            is the distance parameter. For kmeans, it is the number of clusters.

        method : string, optional, default 'leiden'
            Determines which clustering method is run.
            'leiden' - Leiden clustering with modularity optimization
            'leiden_sig' - Leiden clustering with significance optimization
            'louvain' - Louvain clustering with modularity optimization
            'louvain_sig' - Louvain clustering with significance optimization
            'kmeans' - Kmeans clustering
            'dbscan' - DBSCAN clustering
            'hdbscan' - HDBSCAN clustering
            If X is None, cluster assignments are saved in '.adata.obs' with
            key name equal to method + '_clusters'. Otherwise, they are returned.
        """
        if method == "leiden":
            if param is None:
                param = 1
            cl = self.leiden_clustering(X=X, res=param, method="modularity")
        elif method == "leiden_sig":
            if param is None:
                param = 1
            cl = self.leiden_clustering(X=X, res=param, method="significance")
        elif method == "louvain":
            if param is None:
                param = 1
            cl = self.louvain_clustering(X=X, res=param, method="modularity")
        elif method == "louvain_sig":
            if param is None:
                param = 1
            cl = self.louvain_clustering(X=X, res=param, method="significance")
        elif method == "kmeans":
            if param is None:
                param = 6
            cl = self.kmeans_clustering(param, X=X)
        elif method == "hdbscan":
            if param is None:
                param = 25
            cl = self.hdbknn_clustering(npcs=param)
        elif method == "dbscan":
            if param is None:
                param = 0.5
            cl = self.density_clustering(eps=param)
        else:
            cl = None
        return cl

    def louvain_clustering(self, X=None, res=1, method="modularity"):
        """Runs Louvain clustering using the vtraag implementation. Assumes
        that 'louvain' optional dependency is installed.

        Parameters
        ----------
        res - float, optional, default 1
            The resolution parameter which tunes the number of clusters Louvain
            finds.

        method - str, optional, default 'modularity'
            Can be 'modularity' or 'significance', which are two different
            optimizing funcitons in the Louvain algorithm.

        """

        if X is None:
            X = self.adata.obsp["connectivities"]
            save = True
        else:
            if not sp.isspmatrix_csr(X):
                X = sp.csr_matrix(X)
            save = False

        import igraph as ig
        import louvain

        adjacency = X
        sources, targets = adjacency.nonzero()
        weights = adjacency[sources, targets]
        if isinstance(weights, np.matrix):
            weights = weights.A1
        g = ig.Graph(directed=True)
        g.add_vertices(adjacency.shape[0])
        g.add_edges(list(zip(sources, targets)))
        try:
            g.es["weight"] = weights
        except BaseException:
            pass

        if method == "significance":
            cl = louvain.find_partition(g, louvain.SignificanceVertexPartition)
        else:
            cl = louvain.find_partition(
                g, louvain.RBConfigurationVertexPartition, resolution_parameter=res
            )

        if save:
            if method == "modularity":
                self.adata.obs["louvain_clusters"] = pd.Categorical(
                    np.array(cl.membership)
                )
            elif method == "significance":
                self.adata.obs["louvain_sig_clusters"] = pd.Categorical(
                    np.array(cl.membership)
                )
        else:
            return np.array(cl.membership)

    def kmeans_clustering(self, numc, X=None, npcs=25):
        """Performs k-means clustering.

        Parameters
        ----------
        numc - int
            Number of clusters

        npcs - int, optional, default 25
            Number of principal components to use as inpute for k-means
            clustering.

        """

        from sklearn.cluster import KMeans

        if X is None:
            X = self.adata.obsm['X_pca']

        km = KMeans(n_clusters=numc)
        cl = km.fit_predict(Normalizer().fit_transform(X))

        self.adata.obs["kmeans_clusters"] = pd.Categorical(cl)
        return cl, km

    def leiden_clustering(self, X=None, res=1, method="modularity", seed = 0):

        if X is None:
            X = self.adata.obsp["connectivities"]
            save = True
        else:
            if not sp.isspmatrix_csr(X):
                X = sp.csr_matrix(X)
            save = False

        import igraph as ig
        import leidenalg

        adjacency = X
        sources, targets = adjacency.nonzero()
        weights = adjacency[sources, targets]
        if isinstance(weights, np.matrix):
            weights = weights.A1
        g = ig.Graph(directed=True)
        g.add_vertices(adjacency.shape[0])
        g.add_edges(list(zip(sources, targets)))
        try:
            g.es["weight"] = weights
        except BaseException:
            pass

        if method == "significance":
            cl = leidenalg.find_partition(g, leidenalg.SignificanceVertexPartition,seed=seed)
        else:
            cl = leidenalg.find_partition(
                g, leidenalg.RBConfigurationVertexPartition, resolution_parameter=res,seed=seed
            )

        if save:
            if method == "modularity":
                self.adata.obs["leiden_clusters"] = pd.Categorical(
                    np.array(cl.membership)
                )
            elif method == "significance":
                self.adata.obs["leiden_sig_clusters"] = pd.Categorical(
                    np.array(cl.membership)
                )
        else:
            return np.array(cl.membership)

    def hdbknn_clustering(self, X=None, k=None, npcs=15, **kwargs):
        import hdbscan

        if X is None:
            X = self.adata.obsm['X_pca']
            X = Normalizer().fit_transform(X)
            save = True
        else:
            save = False

        if k is None:
            k = self.adata.uns['run_args'].get("k", 20)

        hdb = hdbscan.HDBSCAN(metric="euclidean", **kwargs)

        cl = hdb.fit_predict(X)

        idx0 = np.where(cl != -1)[0]
        idx1 = np.where(cl == -1)[0]
        if idx1.size > 0 and idx0.size > 0:
            xcmap = ut.generate_euclidean_map(X[idx0, :], X[idx1, :])
            knn = np.argsort(xcmap.T, axis=1)[:, :k]
            nnm = np.zeros(xcmap.shape).T
            nnm[
                np.tile(np.arange(knn.shape[0])[:, None], (1, knn.shape[1])).flatten(),
                knn.flatten(),
            ] = 1
            nnmc = np.zeros((nnm.shape[0], cl.max() + 1))
            for i in range(cl.max() + 1):
                nnmc[:, i] = nnm[:, cl[idx0] == i].sum(1)

            cl[idx1] = np.argmax(nnmc, axis=1)

        if save:
            self.adata.obs["hdbscan_clusters"] = pd.Categorical(cl)
        else:
            return cl

    def identify_marker_genes_rf(self, labels=None, clusters=None, n_genes=4000):
        """
        Ranks marker genes for each cluster using a random forest
        classification approach.

        Parameters
        ----------

        labels - numpy.array or str, optional, default None
            Cluster labels to use for marker gene identification.
            Can also be a string corresponding to any of the keys
            in adata.obs.

        clusters - int/string or array-like, default None
            A cluster ID (int or string depending on the labels used)
            or vector corresponding to the specific cluster ID(s) for
            which marker genes will be calculated. If None, marker genes
            will be computed for all clusters, and the result will be written
            to adata.uns.

        n_genes - int, optional, default 4000
            By default, trains the classifier on the top 4000 SAM-weighted
            genes.

        Returns
        -------
        (dictionary of markers for each cluster,
        dictionary of marker scores for each cluster)
        """

        if labels is None:
            try:
                keys = np.array(list(self.adata.obs_keys()))
                lbls = self.get_labels(ut.search_string(keys, "_clusters")[0][0])
            except KeyError:
                print(
                    "Please generate cluster labels first or set the "
                    "'labels' keyword argument."
                )
                return
        elif isinstance(labels, str):
            lbls = self.get_labels(labels)
        else:
            lbls = labels

        from sklearn.ensemble import RandomForestClassifier

        markers = {}
        markers_scores = {}
        if clusters == None:
            lblsu = np.unique(lbls)
        else:
            lblsu = np.unique(clusters)

        indices = np.argsort(-self.adata.var["weights"].values)
        X = self.adata.layers["X_disp"][:, indices[:n_genes]].toarray()
        for K in range(lblsu.size):
            # print(K)
            y = np.zeros(lbls.size)

            y[lbls == lblsu[K]] = 1

            clf = RandomForestClassifier(
                n_estimators=100, max_depth=None, random_state=0
            )

            clf.fit(X, y)

            idx = np.argsort(-clf.feature_importances_)

            markers[lblsu[K]] = self.adata.uns["ranked_genes"][idx]
            markers_scores[lblsu[K]] = clf.feature_importances_[idx]

        if clusters is None:
            if isinstance(labels, str):
                self.adata.uns["rf_" + labels] = markers
            else:
                self.adata.uns["rf"] = markers

        return markers, markers_scores

    def identify_marker_genes_sw(self, labels=None, clusters=None, inplace=True):
        """
        Ranks marker genes for each cluster using partial sums of spatial
        dispersions.

        Parameters
        ----------

        labels - numpy.array or str, optional, default None
            Cluster labels to use for marker gene identification.
            Can also be a string corresponding to any of the keys
            in adata.obs.

        clusters - int/string or array-like, default None
            A cluster ID (int or string depending on the labels used)
            or vector corresponding to the specific cluster ID(s) for
            which marker genes will be calculated. If None, marker genes
            will be computed for all clusters, and the result will be written
            to adata.uns.

        inplace - bool, default True
            If True, returns nothing and places marker scores in `var`

        """

        if labels is None:
            try:
                keys = np.array(list(self.adata.obs_keys()))
                lbls = self.get_labels(ut.search_string(keys, "_clusters")[0][0])
            except KeyError:
                print(
                    "Please generate cluster labels first or set the "
                    "'labels' keyword argument."
                )
                return
        elif isinstance(labels, str):
            lbls = self.get_labels(labels)
        else:
            lbls = labels

        markers_scores = []
        if clusters == None:
            lblsu = np.unique(lbls)
        else:
            lblsu = np.unique(clusters)

        if "X_knn_avg" not in list(self.adata.layers.keys()):
            print("Performing kNN-averaging...")
            self.dispersion_ranking_NN(save_avgs=True)
        l = self.adata.layers["X_knn_avg"]
        m = l.mean(0).A.flatten()
        cells = np.array(list(self.adata.obs_names))
        for K in range(lblsu.size):
            selected = np.where(np.in1d(cells, self.get_cells(lblsu[K], labels)))[0]
            ms = l[selected, :].mean(0).A.flatten()
            lsub = l[selected, :]
            lsub.data[:] = lsub.data ** 2
            ms2 = lsub.mean(0).A.flatten()
            v = ms2 - 2 * ms * m + m ** 2
            wmu = np.zeros(v.size)
            wmu[m > 0] = v[m > 0] / m[m > 0]
            markers_scores.append(wmu)
        A = pd.DataFrame(
            data=np.vstack(markers_scores), index=lblsu, columns=self.adata.var_names
        ).T
        if inplace:
            A.columns = labels + ";;" + A.columns.astype('str').astype('object')
            for Ac in A.columns:
                self.adata.var[Ac]=A[Ac]
        else:
            return A

    def identify_marker_genes_ratio(self, labels=None):
        """
        Ranks marker genes for each cluster using a SAM-weighted
        expression-ratio approach (works quite well).

        Parameters
        ----------

        labels - numpy.array or str, optional, default None
            Cluster labels to use for marker gene identification. If None,
            assumes that one of SAM's clustering algorithms has been run. Can
            be a string (i.e. 'louvain_clusters', 'kmeans_clusters', etc) to
            specify specific cluster labels in adata.obs.

        """
        if labels is None:
            try:
                keys = np.array(list(self.adata.obs_keys()))
                lbls = self.get_labels(ut.search_string(keys, "_clusters")[0][0])
            except KeyError:
                print(
                    "Please generate cluster labels first or set the "
                    "'labels' keyword argument."
                )
                return
        elif isinstance(labels, str):
            lbls = self.get_labels(labels)
        else:
            lbls = labels

        all_gene_names = np.array(list(self.adata.var_names))

        markers = {}

        s = np.array(self.adata.layers["X_disp"].sum(0)).flatten()
        lblsu = np.unique(lbls)
        for i in lblsu:
            d = np.array(self.adata.layers["X_disp"][lbls == i, :].sum(0)).flatten()
            rat = np.zeros(d.size)
            rat[s > 0] = (
                d[s > 0] ** 2 / s[s > 0] * self.adata.var["weights"].values[s > 0]
            )
            x = np.argsort(-rat)
            markers[i] = all_gene_names[x[:]]

        self.adata.uns["marker_genes_ratio"] = markers

        return markers

    def gui(self):
        if 'SamGui' in self.__dict__.keys():
            return self.SamGui.SamPlot
        else:
            from samalg.gui import SAMGUI
            self.SamGui = SAMGUI(self)
            return self.SamGui.SamPlot
            
    def save(self,fn):
        import dill
        if len(fn.split('.pkl')) == 1:
            fn = fn + '.pkl'
        self.path_to_file = fn
        d = {}
        for k in self.__dict__.keys():
            d[k] = self.__dict__[k]
        if 'SamGui' in d.keys():
            del d['SamGui']
        with open(fn,'wb') as f:
            dill.dump(d,f)
            if 'SamGui' in self.__dict__.keys():
                dill.dump(self.SamGui,f)
            
    def load(self,fn):
        import dill
        if len(fn.split('.pkl')) == 1:
            fn = fn + '.pkl'    
        with open(fn,'rb') as f:
            self.__dict__ = dill.load(f)
            try:
                self.__dict__['SamGui'] = dill.load(f)
            except:
                pass;



import pickle
import colorlover as cl
import numpy as np
import scipy.sparse as sp
import ipywidgets as widgets
import plotly.graph_objs as go
from . import SAM
from . import utilities as ut
import pandas as pd

from ipyevents import Event
from ipywidgets import Widget, Layout
import warnings
from numba.core.errors import NumbaWarning
warnings.filterwarnings("ignore", category=NumbaWarning)

__version__ = "0.8.7"


class SAMGUI(object):
    def __init__(self, sam=None,title=None, close_all_widgets=False, default_proj='X_umap',height=600):
        self.ACTION_LOG = []
        
        if close_all_widgets:
            Widget.close_all()
        self.widget_height = height
        self.log('Setting widget height to {}'.format(height))
        
        self.default_proj = default_proj
        self.titles = []
        self.stab = widgets.Tab(layout={"height": "95%", "width": "50%"})
        self.stab.observe(self.on_switch_tabs, "selected_index")

        if sam is not None:
            self.log('Initializing GUI from input SAM object.')
            if not isinstance(sam,list):
                sams=[sam]
            else:
                sams = sam
                sam = sams[0] 
            if title is not None:
                if not isinstance(title,list):
                    titles=[title]
                else:
                    titles = title
                    title = titles[0]
            else:
                titles = [None]*len(sams)
            
            self.init_from_sam(sam)
            if titles[0] is None:
                title = 'Full dataset'
            else:
                title = titles[0]
                
            self.create_plot(0, title)
            items = [self.stab, self.tab]
            self.titles = [title]
            for i in range(1,len(sams)):
                self.sams.append(sams[i])
                self.selected.append(np.ones(sams[i].adata.shape[0]).astype("bool"))
                self.selected_cells.append(np.array(list(sams[i].adata.obs_names)))
                self.active_labels.append(np.zeros(sams[i].adata.shape[0]))
                self.dd_opts.append(["Toggle cluster"])
                self.gene_expressions.append(np.zeros(sams[i].adata.shape[0]))
                self.marker_genes.append(
                    np.array(list(sams[i].adata.var_names))[
                        np.argsort(-sams[i].adata.var["weights"].values)
                    ]
                )
                self.marker_genes_tt.append("Genes ranked by SAM weights.")
                self.ds.append(0)  
                if titles[i] is None:
                    title = 'Full dataset {}'.format(i)
                else:
                    title = titles[i]
                self.titles.append(title)
                self.create_plot(i, title)
                self.update_dropdowns(i)            
        else:
            self.log('Initializing blank GUI.')
            tab = widgets.Tab(layout={"height": "95%", "width": "50%"})
            load_box = self.init_load()
            children = [load_box]
            self.SAM_LOADED = False
            tab.children = children
            tab.set_title(0, "LOAD_SAM")
            self.create_plot(0, "SAM not loaded")
            self.titles = ["SAM not loaded"]
            items = [self.stab, tab]

        self.current_tab = 0
        self.current_sam = sam

        self.SamPlot = widgets.HBox(
            items, layout=Layout(width="auto", height=str(height) + "px")
        )

        x = self.stab.children[self.stab.selected_index]
        f = x.layout.xaxis.range
        x.layout.xaxis.range=(-1,1)
        x.layout.xaxis.range=f        

        f = x.layout.yaxis.range
        x.layout.yaxis.range=(-1,1)
        x.layout.yaxis.range=f
        
    def __setstate__(self,d):       
        self.__init__(sam = d['sams'],title=d['titles'],
                      height=d['widget_height'],default_proj=d['default_proj'])
        
        for k in d.keys():
            self.__dict__[k] = d[k]
        self.__dict__['ds'] = [0]*len(self.sams)

        for i in range(len(self.sams)):
            self.create_plot(i,self.titles[i])
        
    def __getstate__(self):
        d = {}
        exclude = ['SamPlot','children','ps_dict','rb_dict','cs_dict','stab','rs_box','cs_box',
                   'pp_box','out','tab','ds']
        for k in self.__dict__.keys():
            if k not in exclude:
                d[k] = self.__dict__[k]
        return d
        


    def log(self,s):
        self.ACTION_LOG.append(s)
        
    def init_from_sam(self, sam):
        self.current_sam = sam
        tab = widgets.Tab(layout={"height": "95%", "width": "50%"})
        self.load_vars_from_sam(sam)
        self.pp_box = self.init_preprocess()
        self.rs_box = self.init_run_sam()
        self.cs_box = self.init_cs()
        
        self.SAM_LOADED = True
        self.out = widgets.Output(
            layout={"border": "1px solid black", "height": "95%", "width": "100%"}
        )
        children = [self.cs_box, self.rs_box, self.pp_box, self.out]
        self.children = children
        names = ["Interact (0)", "Analysis (0)", "Preprocess (0)", "Output (0)"]
        self.names = ["Interact", "Analysis", "Preprocess", "Output"]
        tab.children = children
        tab.set_trait("selected_index", 0)
        for i in range(len(children)):
            tab.set_title(i, names[i])
        tab.set_trait("selected_index", 0)
        self.tab = tab

    def close_all_tabs(self):
        self.log('Closing all tabs except tab 0.')
        self.stab.set_trait("selected_index", 0)
        I = 1
        while len(self.sams) > 1:
            self.ds[I].close()
            self.stab.children[I].close()
            del self.ds[I]
            del self.sams[I]
            del self.selected[I]
            del self.selected_cells[I]
            del self.active_labels[I]
            del self.dd_opts[I]
            del self.marker_genes[I]
            del self.marker_genes_tt[I]
            del self.gene_expressions[I]
        self.stab.children = (self.stab.children[0],)

    def get_scatter(self, i):
        return self.stab.children[i].data[0]

    def load_vars_from_sam(self, sam):
        self.sams = [sam]
        self.GENE_KEY = ""
        self.selected = [np.zeros(sam.adata.shape[0], dtype="bool")]
        self.selected[0][:] = True
        self.active_labels = [np.zeros(self.selected[0].size, dtype="int")]
        self.dd_opts = [["Toggle cluster"]]
        try:
            self.marker_genes = [
                np.array(list(sam.adata.var_names))[
                    np.argsort(-sam.adata.var["weights"].values)
                ]
            ]
            self.marker_genes_tt = ["Genes ranked by SAM weights."]
        except KeyError:
            self.marker_genes = [np.array(list(sam.adata.var_names))]
            self.marker_genes_tt = ["Unranked genes (SAM not run)."]

        self.selected_cells = [np.array(list(sam.adata.obs_names))]
        self.ds = [0]
        self.gene_expressions = [np.zeros(sam.adata.shape[0])]
        try:
            self.preprocess_args = sam.adata.uns["preprocess_args"].copy()
            for i in list(self.preprocess_args.keys()):
                if isinstance(self.preprocess_args[i], np.ndarray):
                    self.preprocess_args[i] = self.preprocess_args[i][0]
        except:
            self.preprocess_args = {}

        rm = list(invalidArgs(self.sams[0].preprocess_data,self.preprocess_args))
        for k in rm:
            del self.preprocess_args[k]            
            
        self.preprocess_args['thresh'] = self.preprocess_args.get('thresh_low',0.0)
        self.preprocess_args_init = self.preprocess_args.copy()

        try:
            self.run_args = sam.adata.uns["run_args"].copy()
            for i in list(self.run_args.keys()):
                if isinstance(self.run_args[i], np.ndarray):
                    self.run_args[i] = self.run_args[i][0]
        except:
            self.run_args = {}
        
        rm = list(invalidArgs(self.sams[0].run,self.run_args))
        for k in rm:
            del self.run_args[k]
        self.run_args_init = self.run_args.copy()
        


    def create_plot(self, i, title):
        if self.SAM_LOADED:
            projs = list(self.sams[i].adata.obsm.keys())
            if self.default_proj in projs:
                p = self.default_proj
            elif len(projs) > 1:
                p = np.array(projs)[np.where(np.array(projs) != "X_pca")[0][0]]
            else:
                p = "X_pca"

            if p not in projs:
                xdata = []
                ydata = []
            else:
                xdata = self.sams[i].adata.obsm[p][:, 0]
                ydata = self.sams[i].adata.obsm[p][:, 1]

            if i < len(self.stab.children):
                f1 = self.stab.children[i]
                f1.update_layout(autosize=False)
                f1.data = []
            else:
                f1 = go.FigureWidget()

            f1.add_scattergl(x=xdata, y=ydata)

            f1.for_each_trace(self.init_graph)
            f1.update_layout(
                margin_l=0,
                margin_r=0,
                margin_t=40,
                margin_b=0,
                xaxis_ticks="",
                xaxis_showticklabels=False,
                height=self.widget_height * 0.85,
                title="",
                # xaxis_showgrid=False,
                # xaxis_zeroline=False,
                # yaxis_showgrid=False,
                # yaxis_zeroline=False,
                yaxis_ticks="",
                yaxis_showticklabels=False,
                autosize=True,
                dragmode="select",
            )
            f1.update_yaxes(autorange=True)
            f1.update_xaxes(autorange=True)

            f1.data[0].text = list(self.active_labels[i])
            f1.data[0].hoverinfo = "text"
            f1.set_trait(
                "_config",
                {
                    "displayModeBar": True,
                    "scrollZoom": True,
                    "displaylogo": False,
                    "edits": {"titleText": False},
                },
            )

            if i >= len(self.stab.children):
                self.stab.children += (f1,)
                # self.stab.set_trait('selected_index',i)

            if type(self.ds[i]) is int:
                d = Event(source=f1, watched_events=["keydown"])
                self.ds[i] = d
                d.on_dom_event(self.handle_events)

            slider = self.cs_dict["THR"]
            slider.set_trait("min", slider.value)
            slider.set_trait("max", slider.value)

        else:
            f1 = go.FigureWidget()
            f1.add_scattergl(x=[], y=[])
            f1.update_layout(
                margin_l=0,
                margin_r=0,
                margin_t=40,
                margin_b=0,
                height=self.widget_height * 0.85,
                xaxis_ticks="",
                xaxis_showticklabels=False,
                yaxis_ticks="",
                yaxis_showticklabels=False,
                autosize=True,
                dragmode="select",
            )
            f1.set_trait(
                "_config",
                {
                    "displayModeBar": True,
                    "scrollZoom": True,
                    "displaylogo": False,
                    "edits": {"titleText": False},
                },
            )
            self.stab.children += (f1,)

        self.stab.set_title(i, title)

    def handle_events(self, event):
        if event["type"] == "keydown":
            key = event["key"]
            if key == "ArrowRight":
                self.log('Scrolling ranked genes slider one to the right.')
                self.cs_dict["RGENES"].set_trait(
                    "value", self.cs_dict["RGENES"].value + 1
                )
            elif key == "ArrowLeft":
                self.log('Scrolling ranked genes slider one to the left.')
                x = self.cs_dict["RGENES"].value - 1
                if x < 0:
                    x = 0
                self.cs_dict["RGENES"].set_trait("value", x)

            elif key == "Shift":
                self.irm_genes(None)
            elif key == "Enter":
                self.ism_genes(None)
            elif key == "x":
                self.unselect_all(None)
            elif key == "c":
                self.select_all(None)
            elif key == "v":
                self.reset_view(None)
            elif key == "a":
                self.log('Turning off k-nn averaging of gene expression values.')
                self.cs_dict["AVG"].set_trait("value", not self.cs_dict["AVG"].value)

    def close_tab(self, event):
        I = self.stab.selected_index
        if I > 0:
            self.log('Closing tab {}'.format(I))
            self.stab.set_trait("selected_index", I - 1)

            titles = []
            for i in range(len(self.stab.children)):
                titles.append(self.stab.get_title(i))
            del titles[I]

            self.ds[I].close()
            self.stab.children[I].close()
            t = list(self.stab.children)
            del t[I]
            self.stab.children = t
            del self.ds[I]
            del self.sams[I]
            del self.selected[I]
            del self.selected_cells[I]
            del self.gene_expressions[I]
            del self.active_labels[I]
            del self.dd_opts[I]
            del self.marker_genes[I]
            del self.marker_genes_tt[I]

            for i in range(1, len(self.stab.children)):
                self.stab.set_title(i, titles[i])

    """ BEGIN PREPROCESS INIT"""

    def init_preprocess(self):

        pdata = widgets.Button(
            description = 'Preprocess',
            disabled = False,
            tooltip = 'Preprocess data',
            icon = ''
        )
        pdata.on_click(self.preprocess_sam)

        fgenes = widgets.Checkbox(
            value=bool(self.preprocess_args.get("filter_genes", True)),
            description="Filter genes",
        )
        fgenes.observe(self.pp_filtergenes, names="value")
        dfts = widgets.Button(
            description="Set defaults",
            disabled=False,
            tooltip="Set default options",
            icon="",
        )
        dfts.on_click(self.set_pp_defaults)
        
        l1 = widgets.Label("Expr threshold:")
        expr_thr = widgets.FloatSlider(
            value=float(self.preprocess_args.get("thresh", 0.01)),
            min=0,
            max=0.1,
            step=0.005,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="2f",
            style={"description_width": "initial"},
        )
        expr_thr.observe(self.et_update, names="value")

        l2 = widgets.Label("Min expr:")
        min_expr = widgets.FloatSlider(
            value=float(self.preprocess_args.get("min_expression", 1)),
            min=0,
            max=6.0,
            step=0.02,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="1f",
            style={"description_width": "initial"},
        )
        min_expr.observe(self.me_update, names="value")

        init = self.preprocess_args.get("sum_norm", "None")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = "None"
        sumnorm = widgets.Dropdown(
            options=["cell_median", "gene_median", "None"],
            value=init,
            description="Library normalization:",
            disabled=False,
            style={"description_width": "initial"},
        )
        sumnorm.observe(self.sumnorm_submit, "value")

        init = self.preprocess_args.get("norm", "log")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = "None"
        norm = widgets.Dropdown(
            options=["log", "ftt", "None"],
            value=init,
            description="Data normalization:",
            disabled=False,
            style={"description_width": "initial"},
        )
        norm.observe(self.norm_submit, "value")

        load = widgets.Button(
            description="Load data",
            tooltip="Enter the path to the desired data file you wish to "
            "load. Accepted filetypes are .csv (comma), .txt (tab), .h5ad "
            "(AnnData), and .p (pickled SAM dictionary).",
            disabled=False,
        )
        load.on_click(self.load_data)
        load_data = widgets.Text(value="", layout={"width": "100%%"})

        loada = widgets.Button(
            description="Load annotations",
            tooltip="Enter the path to the desired annotations file you wish to "
            "load. Accepted filetypes are .csv (comma) or .txt (tab).",
            disabled=False,
        )
        loada.on_click(self.load_ann)
        load_dataa = widgets.Text(value="", layout={"width": "100%%"})

        loadv = widgets.Button(
            description="Load gene annotations",
            tooltip="Enter the path to the desired gene annotations file "
            "you wish to load. Accepted filetypes are .csv (comma) or .txt (tab).",
            disabled=False,
        )
        loadv.on_click(self.load_vann)
        load_datav = widgets.Text(value="", layout={"width": "100%%"})

        self.ps_dict = {}
        self.ps_dict["PDATA"] = pdata
        self.ps_dict["DFTS"] = dfts
        self.ps_dict["FGENES"] = fgenes
        self.ps_dict["NORM"] = norm
        self.ps_dict["SUMNORM"] = sumnorm
        self.ps_dict["L1"] = l1
        self.ps_dict["EXPR_THR"] = expr_thr
        self.ps_dict["L2"] = l2
        self.ps_dict["MIN_EXPR"] = min_expr
        self.ps_dict["LOAD"] = load
        self.ps_dict["LOAD_DATA"] = load_data
        self.ps_dict["LOADA"] = loada
        self.ps_dict["LOAD_DATAA"] = load_dataa
        self.ps_dict["LOADV"] = loadv
        self.ps_dict["LOAD_DATAV"] = load_datav

        pp = widgets.VBox(
            [   pdata,
                widgets.HBox([dfts, fgenes]),
                norm,
                sumnorm,
                widgets.HBox([l1, expr_thr]),
                widgets.HBox([l2, min_expr]),
                widgets.HBox([load, load_data]),
                widgets.HBox([loada, load_dataa]),
                widgets.HBox([loadv, load_datav]),
            ],
            layout={"height": "95%", "width": "100%"},
        )
        return pp

    def load_ann(self, event):
        path = self.ps_dict["LOAD_DATAA"].value
        self.log('Loading cell annotation file from {}'.format(path))        
        try:
            for i in range(len(self.stab.children)):
                self.sams[i].load_obs_annotations(path)
                self.update_dropdowns(i)
        except:
            with self.out:
                print("Annotation file not found or was improperly formatted.")

    def load_vann(self, event):
        path = self.ps_dict["LOAD_DATAV"].value
        self.log('Loading gene annotation file from {}'.format(path))
        try:
            for i in range(len(self.stab.children)):
                self.sams[i].load_var_annotations(path)
                self.update_dropdowns(i)
        except:
            with self.out:
                print("Annotation file not found or was improperly formatted.")

    def init_load(self):
        load = widgets.Button(
            description="Load data",
            tooltip="Enter the path to the desired data file you wish to "
            "load. Accepted filetypes are .csv (comma), .txt (tab), .h5ad "
            "(AnnData), and .p (pickled SAM dictionary).",
            disabled=False,
        )
        load.on_click(self.load_data)
        load_data = widgets.Text(value="", layout={"width": "100%"})
        return widgets.HBox([load, load_data], layout={"width": "50%"})

    def load_data(self, event):
        try:
            if not self.SAM_LOADED:
                path = self.SamPlot.children[1].children[0].children[1].value
            else:
                path = self.ps_dict["LOAD_DATA"].value
            self.log('Loading data file from {}'.format(path))
            filetype = path.split(".")[-1]
            if filetype == "gz":
                filetype = path.split(".")[-2]

            sam = SAM()
            if filetype == "h5ad" or filetype == "csv":
                sam.load_data(path)
            elif filetype == "p":
                try:
                    sam.load(path)
                except:
                    sam.load_data(path)
            else:
                sam.load_data(path, sep="\t")

            if not self.SAM_LOADED:
                self.SamPlot.children[1].children[0].children[0].close()
                self.SamPlot.children[1].children[0].children[1].close()
                self.SamPlot.children[1].children[0].close()
                self.SamPlot.children[1].close()
                self.init_from_sam(sam)
                self.SamPlot.children = [self.stab, self.tab]
                self.tab.set_trait("selected_index", 1)
            else:
                self.close_all_tabs()
                self.ds[0].close()
                self.ds[0] = 0
                self.load_vars_from_sam(sam)
            self.create_plot(0, "Full dataset")
            if len(self.titles) == 0:
                self.titles = ["Full dataset"]
            else:
                self.titles[0] = "Full dataset"
        except FileNotFoundError:
            0
            # do nothing

    def me_update(self, val):
        self.log('Setting min_expression to be {}'.format(val['new']))
        self.preprocess_args["min_expression"] = val["new"]

    def et_update(self, val):
        self.log('Setting the threshold value to be {}'.format(val['new']))
        self.preprocess_args["thresh"] = val["new"]

    def sumnorm_submit(self, txt):
        self.log('Setting sum_norm value to be {}'.format(txt['new']))
        if txt["new"] == "None":
            self.preprocess_args["sum_norm"] = None
        else:
            self.preprocess_args["sum_norm"] = txt["new"]

    def set_pp_defaults(self, event):
        self.log('Setting preprocessing parameters to defaults.')
        self.preprocess_args = self.preprocess_args_init.copy()
        fgenes = self.ps_dict["FGENES"]  # checkbox
        norm = self.ps_dict["NORM"]
        sumnorm = self.ps_dict["SUMNORM"]
        expr_thr = self.ps_dict["EXPR_THR"]  # expr_threshold
        min_expr = self.ps_dict["MIN_EXPR"]  # min_expr

        init = float(self.preprocess_args.get("min_expression", 1))
        min_expr.set_trait("value", init)

        init = float(self.preprocess_args.get("thresh", 0.01))
        expr_thr.set_trait("value", init)

        init = self.preprocess_args.get("sum_norm", "None")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            sumnorm.set_trait("value", "None")
        else:
            sumnorm.set_trait("value", init)

        init = self.preprocess_args.get("norm", "log")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            norm.set_trait("value", "None")
        else:
            norm.set_trait("value", init)

        init = bool(self.preprocess_args.get("filter_genes", True))
        fgenes.set_trait("value", True)
        self.preprocess_args["filter_genes"] = init

    def norm_submit(self, txt):
        self.log('Setting norm parameter to be {}'.format(txt['new']))
        if txt["new"] == "None":
            self.preprocess_args["norm"] = None
        else:
            self.preprocess_args["norm"] = txt["new"]

    def pp_filtergenes(self, event):
        t = bool(self.preprocess_args.get("filter_genes", True))
        self.log('Setting filter_genes parameter to be {}'.format(not t))
        self.preprocess_args["filter_genes"] = not t

    def preprocess_sam(self, event):
        i = self.stab.selected_index
        self.log('Preprocessing SAM object in tab {}'.format(i))
        self.sams[i].preprocess_data(**self.preprocess_args)

    """ END PREPROCESS INIT"""
    
    """ BEGIN RUN INIT"""

    def init_run_sam(self):
        cpm = widgets.Dropdown(
            options=["UMAP", "t-SNE", "Diffusion map", "Diffusion UMAP"],
            value="UMAP",
            description="",
            layout={"width": "60%"},
        )

        cp = widgets.Button(
            description="Compute projection",
            tooltip="Compute a 2D projection using the selected method in the dropdown menu to the right.",
            layout={"width": "40%"},
        )
        cp.on_click(self.compute_projection)

        clm = widgets.Dropdown(
            options=[
                "Louvain cluster",
                "Density cluster",
                "Hdbscan cluster",
                "Kmeans cluster",
                "Leiden cluster",
            ],
            value="Leiden cluster",
            description="",
            layout={"width": "60%"},
        )
        clm.observe(self.rewire_cluster_slider, "value")
        cl = widgets.Button(
            description="Cluster",
            tooltip="Cluster the data using the selected method in the dropdown menu to the right.",
            layout={"width": "40%"},
        )
        cl.on_click(self.cluster_data)

        l = widgets.Label("Leiden 'res'", layout={"width": "20%"})
        cslider = widgets.FloatSlider(
            value=1,
            min=0.1,
            max=10,
            step=0.1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="2f",
            description="",
            style={"description_width": "initial"},
            layout={"width": "80%"},
        )

        runb = widgets.Button(
            description="Run SAM",
            disabled=False,
            tooltip="Run SAM on the currently selected cells. Enter the title of the new tab in the text box on the right",
            icon="",
        )
        runb.on_click(self.subcluster)
        title = widgets.Text(value="",)
        wpca = widgets.Checkbox(
            value=bool(self.run_args.get("weight_PCs", True)), description="Weight PCs"
        )
        wpca.observe(self.weightpcs)


        spca = widgets.Checkbox(
            value=bool(self.run_args.get("sparse_pca", True)), description="Sparse PCA"
        )
        spca.observe(self.sparsepca)
        

        dfts = widgets.Button(
            description="Set defaults",
            disabled=False,
            tooltip="Set default options",
            icon="",
        )
        dfts.on_click(self.set_run_defaults)

        l1 = widgets.Button(
            description="# genes:",
            disabled=True,
            tooltip="The number of highest-weighted genes to select each iteration of SAM.",
            icon="",
        )
        init = self.run_args.get("n_genes", 3000)
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = 3000
        ngenes = widgets.FloatSlider(
            value=init,
            min=min(100, self.sams[0].adata.shape[1]),
            max=self.sams[0].adata.shape[1],
            step=100,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="5d",
            style={"description_width": "initial"},
        )
        ngenes.observe(self.ngenes_update, names="value")

        l2 = widgets.Button(
            description="# PCs:",
            disabled=True,
            tooltip="The number of principal components to select in PCA.",
            icon="",
        )
        init = self.run_args.get("npcs", 150)
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = 150
        npcs = widgets.FloatSlider(
            value=init,
            min=10,
            max=500,
            step=1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            style={"description_width": "initial"},
        )
        npcs.observe(self.npcs_update, names="value")

        l3 = widgets.Button(
            description="# neighbors:",
            disabled=True,
            tooltip="The number of nearest neighbors to identify for each cell.",
            icon="",
        )
        init = self.run_args.get("k", 20)
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = 20
        knn = widgets.FloatSlider(
            value=init,
            min=5,
            max=200,
            step=1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            style={"description_width": "initial"},
        )
        knn.observe(self.knn_update, names="value")

        l4 = widgets.Button(
            description="num_norm_avg:",
            disabled=True,
            tooltip="The top 'num_norm_avg' dispersions are averaged to determine the "
            "normalization factor when calculating the weights. This prevents "
            "genes with large spatial dispersions from skewing the distribution "
            "of weights.",
            icon="",
        )
        init = self.run_args.get("num_norm_avg", 50)
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = 50
        nna = widgets.FloatSlider(
            value=init,
            min=1,
            max=200,
            step=1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="d",
            style={"description_width": "initial"},
        )
        nna.observe(self.nna_update, names="value")

        init = self.run_args.get("preprocessing", "Normalizer")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = "None"
        norm = widgets.Dropdown(
            options=["StandardScaler", "Normalizer", "None"],
            value=init,
            description="Preprocessing:",
            disabled=False,
            style={"description_width": "initial"},
        )
        norm.observe(self.rnorm_update, "value")

        init = self.run_args.get("distance", "correlation")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = "correlation"
        distance = widgets.Dropdown(
            options=["correlation", "euclidean", "cosine"],
            value=init,
            description="Distance:",
            disabled=False,
            style={"description_width": "initial"},
        )
        distance.observe(self.dist_update, "value")

        init = self.run_args.get("projection", "umap")
        if isinstance(init, np.ndarray):
            init = init[0]
        if init is None:
            init = "umap"
        proj = widgets.Dropdown(
            options=["umap", "tsne", "diff_umap"],
            value=init,
            description="Projection:",
            disabled=False,
            style={"description_width": "initial"},
        )
        proj.observe(self.proj_update, "value")

        self.rb_dict = {}
        self.rb_dict['SPCA'] = spca
        self.rb_dict["RUNB"] = runb
        self.rb_dict["TITLE"] = title
        self.rb_dict["DFTS"] = dfts
        self.rb_dict["WPCA"] = wpca
        self.rb_dict["L3"] = l3
        self.rb_dict["L4"] = l4
        self.rb_dict["NNA"] = nna
        self.rb_dict["KNN"] = knn
        self.rb_dict["NORM"] = norm
        self.rb_dict["DISTANCE"] = distance
        self.rb_dict["PROJ"] = proj
        self.rb_dict["L1"] = l1
        self.rb_dict["L2"] = l2
        self.rb_dict["NGENES"] = ngenes
        self.rb_dict["NPCS"] = npcs
        self.rb_dict["CP"] = cp
        self.rb_dict["CPM"] = cpm
        self.rb_dict["CL"] = cl
        self.rb_dict["CLM"] = clm
        self.rb_dict["L"] = l
        self.rb_dict["CSLIDER"] = cslider

        rs = widgets.VBox(
            [
                widgets.HBox([runb, title]),
                widgets.HBox([dfts, wpca, spca]),
                widgets.HBox([l3, knn]),
                widgets.HBox([l4, nna]),
                norm,
                distance,
                proj,
                widgets.HBox([l1, ngenes]),
                widgets.HBox([l2, npcs]),
                widgets.HBox([cp, cpm]),
                widgets.HBox([cl, clm]),
                widgets.HBox([l, cslider]),
            ],
            layout={"height": "95%", "width": "100%"},
        )

        return rs

    def set_run_defaults(self, event):
        self.log('Setting run parameters to defaults.')
        self.run_args = self.run_args_init.copy()
        # run
        # defaults,wpca
        # ,ngenes
        # ,npcs
        # ,knn
        # ,nna
        # norm,dist,proj
        wpca = self.rb_dict["WPCA"]
        ngenes = self.rb_dict["NGENES"]
        npcs = self.rb_dict["NPCS"]
        knn = self.rb_dict["KNN"]
        nna = self.rb_dict["NNA"]
        rnorm = self.rb_dict["NORM"]
        dist = self.rb_dict["DISTANCE"]
        proj = self.rb_dict["PROJ"]

        init = self.run_args.get("num_norm_avg", 50)
        nna.set_trait("value", init)
        init = self.run_args.get("k", 20)
        knn.set_trait("value", init)

        init = self.run_args.get("npcs", 150)
        if init is None:
            init = 150
        npcs.set_trait("value", init)

        init = self.run_args.get("n_genes", 3000)
        if init is None:
            init = 3000
        ngenes.set_trait("value", init)

        init = self.run_args.get("preprocessing", "Normalizer")
        if init is None:
            init = ""
        rnorm.set_trait("value", init)
        init = self.run_args.get("distance", "correlation")
        dist.set_trait("value", init)
        init = self.run_args.get("projection", "umap")
        proj.set_trait("value", init)

        init = bool(self.run_args.get("weight_PCs", True))
        wpca.set_trait("value", init)
        self.run_args["weight_PCs"] = init

    def sam_weights(self, event):
        self.log('Setting gene rankings to be ordered by the SAM weights.')
        s = self.sams[self.stab.selected_index]
        self.marker_genes[self.stab.selected_index] = np.array(
            list(s.adata.var_names[np.argsort(-s.adata.var["weights"].values)])
        )
        self.marker_genes_tt[self.stab.selected_index] = "Genes ranked by SAM weights."
        self.cs_dict["LGENES"].set_trait(
            "tooltip", self.marker_genes_tt[self.stab.selected_index]
        )

        if self.cs_dict["RGENES"].value != 0:
            self.cs_dict["RGENES"].set_trait("value", 0)
        else:
            self.show_expression(str(self.marker_genes[self.stab.selected_index][0]))

    def weightpcs(self, event):
        t = self.run_args.get("weight_PCs", True)
        self.log('Setting weight_PCs to be {}'.format(not t))
        self.run_args["weight_PCs"] = not t

    def sparsepca(self, event):
        t = self.run_args.get("sparse_pca", True)
        self.log('Setting sparse_pca to be {}'.format(not t))
        self.run_args["sparse_pca"] = not t
        
    def npcs_update(self, val):
        self.log('Setting npcs to be {}'.format(int(val['new'])))
        self.run_args["npcs"] = int(val["new"])

    def nna_update(self, val):
        self.log('Setting num_norm_avg to be {}'.format(int(val['new'])))
        self.run_args["num_norm_avg"] = int(val["new"])

    def knn_update(self, val):
        self.log('Setting k to be {}'.format(int(val['new'])))
        self.run_args["k"] = int(val["new"])

    def ngenes_update(self, val):
        self.log('Setting n_genes to be {}'.format(int(val['new'])))
        self.run_args["n_genes"] = int(val["new"])

    def rnorm_update(self, txt):
        self.log('Setting preprocessing to be {}'.format(txt['new']))
        if txt["new"] == "None":
            self.run_args["preprocessing"] = None
        else:
            self.run_args["preprocessing"] = txt["new"]

    def proj_update(self, txt):
        self.log('Setting projection to be {}'.format(txt['new']))
        self.run_args["projection"] = txt["new"]

    def dist_update(self, txt):
        self.log('Setting distance to be {}'.format(txt['new']))
        self.run_args["distance"] = txt["new"]

    def subcluster(self, event):

        execute = False
        i = self.stab.selected_index

        selected = self.selected[i]
        selected_cells = self.selected_cells[i]

        sam = self.sams[i]
        if not np.all(selected) and selected.sum() > 0:
            sam_subcluster = SAM(counts=sam.adata[selected_cells, :].copy())
            sam_subcluster.adata_raw = sam.adata_raw[selected_cells].copy()

            sam_subcluster.adata_raw.obs = sam_subcluster.adata.obs.copy()
            sam_subcluster.adata_raw.obsm = sam_subcluster.adata.obsm.copy()

            #sam_subcluster.preprocess_data(**self.preprocess_args)
            
            self.out.clear_output()
            with self.out:
                sam_subcluster.run(**self.run_args)
            self.sams.append(sam_subcluster)
            self.selected.append(np.ones(sam_subcluster.adata.shape[0]).astype("bool"))
            self.selected_cells.append(np.array(list(sam_subcluster.adata.obs_names)))
            self.active_labels.append(np.zeros(sam_subcluster.adata.shape[0]))
            self.dd_opts.append(["Toggle cluster"])
            self.gene_expressions.append(np.zeros(sam_subcluster.adata.shape[0]))
            self.marker_genes.append(
                np.array(list(sam_subcluster.adata.var_names))[
                    np.argsort(-sam_subcluster.adata.var["weights"].values)
                ]
            )
            self.marker_genes_tt.append("Genes ranked by SAM weights.")
            self.ds.append(0)
            i = len(self.sams) - 1
            execute = True

        elif np.all(selected):
            sam.preprocess_data(**self.preprocess_args)

            self.out.clear_output()
            with self.out:
                sam.run(**self.run_args)
            self.marker_genes[i] = np.array(list(sam.adata.var_names))[
                np.argsort(-sam.adata.var["weights"].values)
            ]
            self.marker_genes_tt[i] = "Genes ranked by SAM weights."
            execute = True
            self.current_sam = sam

        if execute:
            self.log(('Subclustering SAM object in tab {}'.format(i),selected_cells))
            title = self.rb_dict["TITLE"].value
            if title == "" and i == 0:
                title = "Full dataset"
            elif title == "":
                title = "Subcluster " + str(i)
            self.create_plot(i, title)
            try:
                self.titles[i] = title
            except:
                self.titles.append(title)
                
            self.update_dropdowns(i)

    """ END RUN INIT"""

    """ BEGIN CS INIT"""

    def init_cs(self):
        initl = list(self.sams[0].adata.obsm.keys())
        initl = [""] + initl
        if self.default_proj in initl:
            init = self.default_proj
        else:
            init = initl[0]

        dpm = widgets.Dropdown(
            options=initl, value=init, description="", layout={"width": "60%"}
        )
        dp = widgets.Button(
            description="Display projection",
            tooltip="Display the 2D projection selected in the dropdown menu to the right.",
            layout={"width": "40%"},
        )
        dp.on_click(self.display_projection)

        initl = list(self.sams[0].adata.obs.keys())
        initl = [""] + initl
        dam = widgets.Dropdown(
            options=initl, value=initl[0], description="", layout={"width": "60%"}
        )
        da = widgets.Button(
            description="Display cell ann.",
            tooltip="Overlay the annotations selected in the dropdown menu to the right.",
            layout={"width": "40%"},
        )
        da.on_click(self.display_annotation)

        dv = widgets.Button(
            description="Display gene ann.",
            tooltip="Include the 'var' annotation for the corresponding gene in the title of the gene expression plots.",
            layout={"width": "40%"},
        )
        dv.on_click(self.display_var_annotation)

        initl = list(self.sams[0].adata.var.keys())
        initl = [""] + initl
        dvm = widgets.Dropdown(
            options=initl, value=initl[0], description="", layout={"width": "60%"}
        )

        irm = widgets.Button(
            description="Find markers (RF)",
            tooltip="Identify marker genes of selected cells using a random forest classifier.",
            layout={"width": "34%"},
        )
        irm.on_click(self.irm_genes)
        ism = widgets.Button(
            description="Find markers (SW)",
            tooltip="Identify marker genes of selected cells by finding genes with large SAM weights among the selected cells.",
            layout={"width": "33%"},
        )
        ism.on_click(self.ism_genes)

        sws = widgets.Button(
            description="SAM ranking",
            tooltip="Ranks genes by SAM weights.",
            layout={"width": "33%"},
        )
        sws.on_click(self.sam_weights)

        us = widgets.Button(
            description="Unselect all (x)",
            tooltip="Unselect all cells. Pressing 'x' on the keyboard while hovering over the scatter plot will do the same thing.",
            layout={"width": "34%"},
        )
        usa = widgets.Button(
            description="Select all (c)",
            tooltip="Select all cells. Pressing 'c' on the keyboard while hovering over the scatter plot will do the same thing.",
            layout={"width": "33%"},
        )
        res = widgets.Button(
            description="Reset view (v)",
            tooltip="Resets the current plot to the defautl view. Pressing 'v' on the keyboard while hovering over the scatter plot"
            " will do the same thing.",
            layout={"width": "33%"},
        )
        us.on_click(self.unselect_all)
        usa.on_click(self.select_all)
        res.on_click(self.reset_view)

        avg = widgets.Checkbox(
            indent=False, value=True, description="Avg expr", layout={"width": "20%"}
        )

        log = widgets.Checkbox(
            indent=False, value=False, description="Log cbar", layout={"width": "20%"}
        )

        show_colorbar = widgets.Checkbox(
            indent=False, value=False, description="Show cbar", layout={"width": "20%"}
        )
        lann = widgets.Button(
            description="Annotate",
            tooltip=(
                "Enter the key of the 'obs' annotation vector you wish to modify."
                " If the key does not exist, a new vector in 'obs' will be created. Enter the"
                " label you want to annotate the selected cells with in the right."
            ),
            disabled=True,
            layout={"width": "20%"},
        )
        anno_name = widgets.Text(
            value="", placeholder="Annotation column name", layout={"width": "40%"}
        )
        anno = widgets.Text(value="", placeholder="Annotation", layout={"width": "40%"})
        anno.on_submit(self.annotate_pop)

        lgsm = widgets.Button(
            description="Get similar genes",
            tooltip="Rank genes in order of decreasing spatially averaged correlation with respect to the input gene.",
            disabled=True,
            layout={"width": "30%"},
        )
        gsm = widgets.Text(
            value="",
            placeholder=self.sams[0].adata.var_names[0],
            layout={"width": "50%"},
        )
        gsm.on_submit(self.get_similar_genes)
        lshg = widgets.Button(
            description="Show expressions",
            tooltip="Overlay the input gene expressions over the scatter plot. Only a substring of the gene need be provided.",
            disabled=True,
            layout={"width": "30%"},
        )
        shg = widgets.Text(
            value="",
            placeholder=self.sams[0].adata.var_names[0],
            layout={"width": "50%"},
        )
        shg.on_submit(self.show_expression)

        lgenes = widgets.Button(
            description="Ranked genes",
            tooltip="Genes ranked by SAM weights.",
            disabled=True,
            layout={"width": "30%"},
        )
        rgenes = widgets.IntSlider(
            value=0,
            min=0,
            max=200,
            step=1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            description="",
            style={"description_width": "initial"},
            layout={"width": "70%"},
        )
        rgenes.observe(self.gene_update, "value")

        lthr = widgets.Button(
            description="Threshold expr",
            tooltip="Select genes with greater than the chosen expression threshold.",
            disabled=True,
            layout={"width": "30%"},
        )
        thr = widgets.FloatSlider(
            value=0,
            min=0,
            max=0,
            step=0.05,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            description="",
            style={"description_width": "initial"},
            layout={"width": "70%"},
        )
        thr.observe(self.threshold_selection, "value")

        lsf = widgets.Button(
            description="Save",
            tooltip="Save the current SAM object. Filenames should end with .h5ad.",
            disabled=True,
            layout={"width": "30%"},
        )

        sf = widgets.Text(value="", layout={"width": "70%"})
        sf.on_submit(self.save_data)

        close = widgets.Button(
            description="Close current tab",
            tooltip="Closes the currently open subclustering tab.",
            disabled=False,
        )  # layout={'width':'30%'})
        close.on_click(self.close_tab)

        hotkeys = widgets.Button(
            description="ReadMe",
            tooltip="""While hovering over the scatter plot, the following keyboard inputs are available:
                - Left/Right arrows: Scroll through & display ranked genes
                - Shift: Random Forest classifier marker gene identification of selected cells
                - Enter: SAM weights-based approach to marker gene identification of selected cells
                - x: Unselect all cells
                - c: Select all cells
                - v: Reset view
                - a: Toggle expression spatial averaging ON/OFF

                Also note that clicking a cell will select/unselect all cells sharing its label.""",
            disabled=True,
        )  # layout={'width':'30%'})

        amslider = widgets.Button(
            description="Opacity",
            tooltip="Changes the opacities of unselected points in the current plot.",
            disabled=True,
            layout={"width": "30%"},
        )
        aslider = widgets.FloatSlider(
            value=0.05,
            min=0.0,
            max=1,
            step=0.01,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="2f",
            description="",
            style={"description_width": "initial"},
            layout={"width": "70%"},
        )
        aslider.observe(self.change_alpha, "value")

        lmslider = widgets.Button(
            description="Marker size",
            tooltip="Changes the marker sizes in the current plot.",
            disabled=True,
            layout={"width": "30%"},
        )
        mslider = widgets.FloatSlider(
            value=5,
            min=0.1,
            max=20,
            step=0.1,
            disabled=False,
            continuous_update=False,
            orientation="horizontal",
            readout=True,
            readout_format="2f",
            description="",
            style={"description_width": "initial"},
            layout={"width": "70%"},
        )
        mslider.observe(self.change_msize, "value")

        acc = widgets.Dropdown(value="Toggle cluster", options=["Toggle cluster"])
        acc.observe(self.pick_cells_dd, "value")

        self.cs_dict = {}
        self.cs_dict["DP"] = dp
        self.cs_dict["DPM"] = dpm
        self.cs_dict["DA"] = da
        self.cs_dict["DAM"] = dam
        self.cs_dict["ACC"] = acc
        self.cs_dict["DV"] = dv
        self.cs_dict["DVM"] = dvm
        self.cs_dict["IRM"] = irm
        self.cs_dict["ISM"] = ism
        self.cs_dict["SWS"] = sws
        self.cs_dict["US"] = us
        self.cs_dict["USA"] = usa
        self.cs_dict["RES"] = res
        self.cs_dict["LANN"] = lann
        self.cs_dict["ANNO_NAME"] = anno_name
        self.cs_dict["ANNO"] = anno
        self.cs_dict["SHOW_COLORBAR"] = show_colorbar
        self.cs_dict["LGSM"] = lgsm
        self.cs_dict["GSM"] = gsm
        self.cs_dict["AVG"] = avg
        self.cs_dict["LSHG"] = lshg
        self.cs_dict["SHG"] = shg
        self.cs_dict["LOG"] = log
        self.cs_dict["LGENES"] = lgenes
        self.cs_dict["RGENES"] = rgenes
        self.cs_dict["LTHR"] = lthr
        self.cs_dict["THR"] = thr
        self.cs_dict["LSF"] = lsf
        self.cs_dict["SF"] = sf
        self.cs_dict["AMSLIDER"] = amslider
        self.cs_dict["ASLIDER"] = aslider
        self.cs_dict["LMSLIDER"] = lmslider
        self.cs_dict["MSLIDER"] = mslider
        self.cs_dict["CLOSE"] = close
        self.cs_dict["HOTKEYS"] = hotkeys

        return widgets.VBox(
            [
                widgets.HBox([dp, dpm]),
                widgets.HBox([da, dam, acc]),
                widgets.HBox([dv, dvm]),
                widgets.HBox([irm, ism, sws]),
                widgets.HBox([us, usa, res]),
                widgets.HBox([lann, anno_name, anno,]),
                widgets.HBox([lgsm, gsm,]),
                widgets.HBox([lshg, shg,]),
                widgets.HBox([lgenes, rgenes]),
                widgets.HBox([lthr, thr]),
                widgets.HBox([amslider, aslider]),
                widgets.HBox([show_colorbar, avg, log]),
                widgets.HBox([lmslider, mslider]),
                widgets.HBox([lsf, sf]),
                widgets.HBox([close, hotkeys]),
            ],
            layout={"height": "95%", "width": "100%"},
        )

    def reset_view(self, event):
        i = self.stab.selected_index
        self.log('Resetting view for tab {}'.format(i))        
        self.create_plot(i, self.stab.get_title(i))
        self.marker_genes[i] = np.array(list(self.sams[i].adata.var_names))[
            np.argsort(-self.sams[i].adata.var["weights"].values)
        ]
        self.marker_genes_tt[i] = "Genes ranked by SAM weights."
        self.cs_dict["LGENES"].set_trait("tooltip", self.marker_genes_tt[i])

    def save_data(self, path):
        path = path.value
        self.log('Saving data from tab {} to {}'.format(self.stab.selected_index,path))
        if path != "":
            if path.split(".")[-1] == "h5ad":
                s = self.sams[self.stab.selected_index]
                s.save_anndata(path)
            elif (
                path.split(".")[-1] == "png"
                or path.split(".")[-1] == "pdf"
                or path.split(".")[-1] == "eps"
                or path.split(".")[-1] == "jpg"
            ):
                if len(path.split("/")) > 1:
                    ut.create_folder("/".join(path.split("/")[:-1]))
                self.stab.children[self.stab.selected_index].write_image(path)

    def show_expression(self, text):
        if type(text) is not str:
            gene = text.value
        else:
            gene = text

        s = self.sams[self.stab.selected_index]
        if gene != "":

            try:
                if not (gene in s.adata.var_names):
                    genes = ut.search_string(
                        np.array(list(s.adata.var_names)), gene, case_sensitive=True
                    )[0]
                    if genes is not -1:
                        gene = genes[0]

                if self.cs_dict["AVG"].value:
                    if "X_knn_avg" not in s.adata.layers.keys():
                        s.dispersion_ranking_NN(save_avgs=True)

                    x = s.adata[:, gene].layers["X_knn_avg"]
                    if sp.issparse(x):
                        a = x.A.flatten()
                    else:
                        a = x.flatten()
                else:
                    x = s.adata[:, gene][s.adata.obs_names, :].X
                    if sp.issparse(x):
                        a = x.A.flatten()
                    else:
                        a = x.flatten()

                if self.GENE_KEY != "":
                    title = gene + "; " + str(s.adata.var[self.GENE_KEY].T[gene])
                else:
                    title = gene

                # self.select_all(None)
                self.update_colors_expr(a, title)
                self.log('Showing gene expression values for {} in tab {}'.format(title,self.stab.selected_index))
            except:
                if self.cs_dict["AVG"].value:
                    if not (gene in s.adata.var_names):
                        with self.out:
                            print(
                                "X_knn_avg does not exist. Either run"
                                " sam.dispersion_ranking_NN() or uncheck"
                                " `avg` in the control panel."
                            )
                elif not (gene in s.adata.var_names):
                    with self.out:
                        print(
                            "Gene not found. Check your spelling and " "capitalization."
                        )

    def get_similar_genes(self, txt):
        gene = txt.value
        if gene != "":
            s = self.sams[self.stab.selected_index]
            if "X_knn_avg" not in s.adata.layers.keys():
                s.dispersion_ranking_NN()

            genes = ut.search_string(
                np.array(list(s.adata.var_names)), gene, case_sensitive=True
            )[0]
            if genes is not -1:
                gene = genes[0]
            else:
                return
                # quit

            markers = ut.find_corr_genes(s, gene).flatten()
            _, i = np.unique(markers, return_index=True)
            markers = markers[np.sort(i)]
            self.marker_genes[self.stab.selected_index] = markers

            self.cs_dict["RGENES"].set_trait("value", 0)
            self.marker_genes_tt[self.stab.selected_index] = (
                "Ranked genes from most to least spatially correlated with "
                + gene
                + "."
            )
            self.cs_dict["LGENES"].set_trait(
                "tooltip", self.marker_genes_tt[self.stab.selected_index]
            )
            self.log('Getting genes with similar expression patterns to {}'.format(gene))
            self.show_expression(str(gene))
            

    def annotate_pop(self, text):
        text = text.value
        text_name = self.cs_dict["ANNO_NAME"].value
        selected = self.selected[self.stab.selected_index]
        selected_cells = self.selected_cells[self.stab.selected_index]
        if text != "" and text_name != "" and selected.sum() != selected.size:
            self.log(('Annotating cells with the {} label in the {} key.'.format(text,text_name),selected_cells))
            for it, s in enumerate(self.sams):
                try:
                    x1 = np.array(list(s.adata.obs_names))
    
                    if text_name in list(s.adata.obs.keys()):
                        a = s.get_labels(text_name).copy().astype("<U100")
                        a[np.in1d(x1, selected_cells)] = text
                        s.adata.obs[text_name] = pd.Categorical(a)
    
                    else:
                        a = np.zeros(s.adata.shape[0], dtype="<U100")
                        a[:] = ""
                        a[np.in1d(x1, selected_cells)] = text
                        s.adata.obs[text_name] = pd.Categorical(a)
                except:
                    pass
                
                self.update_dropdowns(it)

    def unselect_all(self, event):
        self.log('Unselecting all cells.')
        self.selected[self.stab.selected_index][:] = False
        self.selected_cells[self.stab.selected_index] = []
        self.stab.children[self.stab.selected_index].data[0].selectedpoints = np.array(
            []
        )
        self.stab.children[self.stab.selected_index].data[0].unselected = {
            "marker": {"opacity": self.cs_dict["ASLIDER"].value}
        }

    def select_all(self, event):
        self.log('Selecting all cells.')
        self.selected[self.stab.selected_index][:] = True
        self.selected_cells[self.stab.selected_index] = np.array(
            list(self.sams[self.stab.selected_index].adata.obs_names)
        )
        self.stab.children[self.stab.selected_index].data[0].selectedpoints = list(
            np.arange(self.selected[self.stab.selected_index].size)
        )
        self.stab.children[self.stab.selected_index].data[0].marker.opacity = 1

    def ism_genes(self, event):
        selected = self.selected[self.stab.selected_index]
        s = self.sams[self.stab.selected_index]
        if not np.all(selected) and selected.sum() > 0:
            self.log(('Calculating marker genes using SW method in tab {}'.format(self.stab.selected_index),self.selected_cells[self.stab.selected_index]))           

            if "X_knn_avg" not in s.adata.layers.keys():
                s.dispersion_ranking_NN(save_avgs=True)                
            l = s.adata.layers["X_knn_avg"]
            m = l.mean(0).A.flatten()
            ms = l[selected, :].mean(0).A.flatten()
            lsub = l[selected, :]
            lsub.data[:] = lsub.data ** 2
            ms2 = lsub.mean(0).A.flatten()
            v = ms2 - 2 * ms * m + m ** 2
            wmu = np.zeros(v.size)
            wmu[m > 0] = v[m > 0] / m[m > 0]

            self.marker_genes[self.stab.selected_index] = np.array(
                list(s.adata.var_names[np.argsort(-wmu)])
            )
            if self.cs_dict["RGENES"].value != 0:
                self.cs_dict["RGENES"].set_trait("value", 0)
            else:
                self.show_expression(
                    str(self.marker_genes[self.stab.selected_index][0])
                )
            self.marker_genes_tt[
                self.stab.selected_index
            ] = "Ranked genes according to spatial SAM weights among selected cells."
            self.cs_dict["LGENES"].set_trait(
                "tooltip", self.marker_genes_tt[self.stab.selected_index]
            )

    def irm_genes(self, event):
        selected = self.selected[self.stab.selected_index]
        s = self.sams[self.stab.selected_index]
        if not np.all(selected) and selected.sum() > 0:
            self.log(('Calculating marker genes using RF method in tab {}'.format(self.stab.selected_index),self.selected_cells[self.stab.selected_index]))            
            a = np.zeros(s.adata.shape[0])
            a[selected] = 1
            markers, _ = s.identify_marker_genes_rf(labels=a, clusters=1)
            self.marker_genes[self.stab.selected_index] = markers[1]
            if self.cs_dict["RGENES"].value != 0:
                self.cs_dict["RGENES"].set_trait("value", 0)
            else:
                self.show_expression(
                    str(self.marker_genes[self.stab.selected_index][0])
                )
            self.marker_genes_tt[
                self.stab.selected_index
            ] = "Ranked genes according to random forest classification."
            self.cs_dict["LGENES"].set_trait(
                "tooltip", self.marker_genes_tt[self.stab.selected_index]
            )

    def gene_update(self, val):
        val = val["new"]
        markers = self.marker_genes[self.stab.selected_index]

        if int(val) < markers.size:
            gene = markers[int(val)]
        else:
            gene = markers[-1]
        self.cs_dict["SHG"].set_trait("value", gene)
        self.show_expression(str(gene))

    def update_dropdowns(self, i):        
        s = self.sams[i]
        self.cs_dict["DPM"].options = [""] + list(s.adata.obsm.keys())
        self.cs_dict["DAM"].options = [""] + list(s.adata.obs.keys())
        self.cs_dict["DVM"].options = [""] + list(s.adata.var.keys())
        self.cs_dict["ACC"].options = self.dd_opts[i]

    def on_switch_tabs(self, event):
        self.log('Switching to tab {}'.format(self.stab.selected_index))
        self.update_dropdowns(self.stab.selected_index)
        self.cs_dict["LGENES"].set_trait(
            "tooltip", self.marker_genes_tt[self.stab.selected_index]
        )
        self.current_tab = self.stab.selected_index
        self.current_sam = self.sams[self.current_tab]
        for i in range(len(self.children)):
            self.tab.set_title(
                i, self.names[i] + " (" + str(self.stab.selected_index) + ")"
            )

        x = self.stab.children[self.stab.selected_index]
        f = x.layout.xaxis.range
        x.layout.xaxis.range=(-1,1)
        x.layout.xaxis.range=f        

        f = x.layout.yaxis.range
        x.layout.yaxis.range=(-1,1)
        x.layout.yaxis.range=f

            

    def display_var_annotation(self, event):
        self.GENE_KEY = self.cs_dict["DVM"].value
        self.log('Setting the var annotations to be from key {}'.format(self.GENE_KEY))

    def display_annotation(self, event):
        key = self.cs_dict["DAM"].value
        self.log('Displaying cell annotations from obs key {}'.format(key))
        if key != "":
            labels = np.array(list(self.sams[self.stab.selected_index].get_labels(key)))

            self.active_labels[self.stab.selected_index] = labels
            self.update_colors_anno(labels)

    def update_colors_expr(self, a, title):
        if self.cs_dict["LOG"].value:
            a = np.log2(a + 1)
        if self.cs_dict["SHOW_COLORBAR"].value:
            showscale = True
        else:
            showscale = False
        self.gene_expressions[self.stab.selected_index] = a

        f1 = self.stab.children[self.stab.selected_index]
        f1.update_traces(
            marker=dict(
                color=a,
                colorscale="spectral",
                reversescale=True,
                showscale=showscale,
                colorbar_ticks="outside",
                colorbar_tickmode="auto",
                colorbar_title="",
                opacity=1,
            )
        )

        f1.update_layout(hovermode="closest", title=title)
        self.stab.children[self.stab.selected_index].data[0].text = list(a)
        self.stab.children[self.stab.selected_index].data[0].hoverinfo = "text"

        slider = self.cs_dict["THR"]
        slider.set_trait("min", 0)
        slider.set_trait("max", a.max() + (a.max() - a.min()) / 100)
        slider.set_trait("step", (a.max() - a.min()) / 100)
        slider.set_trait("value", 0)

    def update_colors_anno(self, labels):
        nlabels = np.unique(labels).size
        title = self.cs_dict["DAM"].value.split("_clusters")[0]

        if issubclass(labels.dtype.type, np.number):
            if issubclass(labels.dtype.type, np.float):
                self.update_colors_expr(labels, title)
                return

        if nlabels == 1:
            x = "spectral"
        elif nlabels <= 2:
            x = cl.scales["3"]["qual"]["Paired"][:nlabels]
        elif nlabels <= 11:
            nlabels = str(nlabels)
            x = cl.scales[nlabels]["qual"]["Paired"]
        else:
            nlabels = "11"
            x = cl.scales[nlabels]["qual"]["Paired"]
            x = cl.interp(x, int(nlabels))

        lbls, inv = np.unique(labels, return_inverse=True)
        self.dd_opts[self.stab.selected_index] = ["Toggle cluster"] + list(lbls)
        self.cs_dict["ACC"].options = self.dd_opts[self.stab.selected_index]

        if issubclass(labels.dtype.type, np.character):
            tickvals = np.arange(lbls.size)
            ticktext = list(lbls)
        else:
            idx = np.round(np.linspace(0, len(lbls) - 1, 6)).astype(int)
            tickvals = list(idx)
            ticktext = lbls[tickvals]

        if self.cs_dict["SHOW_COLORBAR"].value:
            showscale = True
        else:
            showscale = False

        f1 = self.stab.children[self.stab.selected_index]
        f1.update_traces(
            marker=dict(
                color=inv,
                colorscale=x,
                colorbar_ticks="outside",
                colorbar_tickmode="array",
                colorbar_title="",
                colorbar_tickvals=tickvals,
                showscale=showscale,
                colorbar_ticktext=ticktext,
                opacity=1,
            )
        )
        self.stab.children[self.stab.selected_index].layout.title = title
        self.stab.children[self.stab.selected_index].data[0].text = list(lbls[inv])
        self.stab.children[self.stab.selected_index].data[0].hoverinfo = "text"
        self.stab.children[self.stab.selected_index].layout.hovermode = "closest"

        slider = self.cs_dict["THR"]
        slider.set_trait("min", slider.value)
        slider.set_trait("max", slider.value)

    def threshold_selection(self, val):
        val = val["new"]
        self.log('Setting gene expression selection threshold to be {}'.format(val))
        s = self.sams[self.stab.selected_index]

        self.selected[self.stab.selected_index][:] = False
        self.selected[self.stab.selected_index][
            self.gene_expressions[self.stab.selected_index] >= val
        ] = True
        self.selected_cells[self.stab.selected_index] = np.array(
            list(s.adata.obs_names)
        )[self.selected[self.stab.selected_index]]
        self.stab.children[self.stab.selected_index].data[0].selectedpoints = np.where(
            self.selected[self.stab.selected_index]
        )[0]

    def rewire_cluster_slider(self, event):
        x = self.rb_dict["CSLIDER"]
        l = self.rb_dict["L"]
        val = event["new"]
        if val == "Kmeans cluster":
            x.set_trait("min", 2)
            x.set_trait("max", 200)
            x.set_trait("value", 6)
            x.set_trait("step", 1)
            l.set_trait("value", "Kmeans 'k'")

        elif val == "Louvain cluster":
            x.set_trait("min", 0.1)
            x.set_trait("max", 10)
            x.set_trait("value", 1)
            x.set_trait("step", 0.1)
            l.set_trait("value", "Louvain 'res'")

        elif val == "Density cluster":
            x.set_trait("min", 0.1)
            x.set_trait("max", 2)
            x.set_trait("value", 0.5)
            x.set_trait("step", 0.01)
            l.set_trait("value", "Density 'eps'")

        elif val == "Hdbscan cluster":
            l.set_trait("value", "Hdbscan 'N/A'")

        elif val == "Leiden cluster":
            x.set_trait("min", 0.1)
            x.set_trait("max", 10)
            x.set_trait("value", 1)
            x.set_trait("step", 0.1)
            l.set_trait("value", "Leiden 'res'")

    def cluster_data(self, event):
        s = self.sams[self.stab.selected_index]
        val = self.rb_dict["CLM"].value
        eps = self.rb_dict["CSLIDER"].value
        self.log('Clustering data using the {} method with parameter {} in tab {}'.format(val,eps,self.stab.selected_index))
        if val == "Kmeans cluster":
            s.kmeans_clustering(int(eps))

        elif val == "Louvain cluster":
            s.louvain_clustering(res=eps)

        elif val == "Density cluster":
            s.density_clustering(eps=eps)

        elif val == "Hdbscan cluster":
            s.hdbknn_clustering(k=int(eps))

        elif val == "Leiden cluster":
            s.leiden_clustering(res=eps)

        self.cs_dict["DAM"].options = [""] + list(s.adata.obs.keys())

    def display_projection(self, event):
        s = self.sams[self.stab.selected_index]
        key = self.cs_dict["DPM"].value
        self.log('Displaying projection from obsm key {} in tab {}'.format(key,self.stab.selected_index))
        if key != "":
            X = s.adata.obsm[key][:, :2]
            self.stab.children[self.stab.selected_index].data[0]["x"] = X[:, 0]
            self.stab.children[self.stab.selected_index].data[0]["y"] = X[:, 1]

    def compute_projection(self, event):
        i = self.stab.selected_index
        s = self.sams[i]
        val = self.rb_dict["CPM"].value
        if val == "UMAP":
            s.run_umap()
        elif val == "t-SNE":
            s.run_tsne()
        elif val == "Diffusion map":
            s.run_diff_map()
        elif val == "Diffusion UMAP":
            s.run_diff_umap()
        self.log('Computing 2D projection using the {} method'.format(val))
        self.update_dropdowns(i)

    """ END CS INIT"""

    def select(self, trace, points, selector):
        if np.array(points.point_inds).size > 0:
            self.selected[self.stab.selected_index][np.array(points.point_inds)] = True
            self.selected_cells[self.stab.selected_index] = np.array(
                list(
                    self.sams[self.stab.selected_index].adata.obs_names[
                        self.selected[self.stab.selected_index]
                    ]
                )
            )
        trace.selectedpoints = list(
            np.where(self.selected[self.stab.selected_index])[0]
        )
        trace.unselected = {"marker": {"opacity": self.cs_dict["ASLIDER"].value}}

    def pick_cells_dd(self, txt):
        if txt["new"] != "Toggle cluster":
            al = str(txt["new"])
            sel = self.selected[self.stab.selected_index]
            als = self.active_labels[self.stab.selected_index].astype("str")
            ratio = sel[als == al].sum() / sel[als == al].size
            if ratio >= 0.5:
                sel[als == al] = False
            else:
                sel[als == al] = True

            self.selected_cells[self.stab.selected_index] = np.array(
                list(self.sams[self.stab.selected_index].adata.obs_names[sel])
            )

            self.stab.children[self.stab.selected_index].data[0].selectedpoints = list(
                np.where(sel)[0]
            )
            self.log('Toggling cluster {} in tab {}'.format(al,self.stab.selected_index))

            self.cs_dict["ACC"].value = "Toggle cluster"

    def pick_cells(self, trace, points, selector):
        tf = self.selected[self.stab.selected_index][points.point_inds[0]]
        als = self.active_labels[self.stab.selected_index]
        al = als[points.point_inds[0]]

        if tf:
            self.selected[self.stab.selected_index][als == al] = False
        else:
            self.selected[self.stab.selected_index][als == al] = True

        self.selected_cells[self.stab.selected_index] = np.array(
            list(
                self.sams[self.stab.selected_index].adata.obs_names[
                    self.selected[self.stab.selected_index]
                ]
            )
        )
        trace.selectedpoints = list(
            np.where(self.selected[self.stab.selected_index])[0]
        )

    def change_alpha(self, val):
        # m = self.stab.children[self.stab.selected_index].data[0].marker.copy()
        # m['opacity'] = val['new']
        self.stab.children[self.stab.selected_index].data[0].unselected = {
            "marker": {"opacity": val["new"]}
        }
        self.log('Changing alpha (opacity) to be {}'.format(val['new']))

    def change_msize(self, val):
        val = val["new"]
        markers = self.stab.children[self.stab.selected_index].data[0].marker
        markers.size = val
        self.log('Changing marker size to be {}'.format(val))

    def init_graph(self, trace):
        trace.mode = "markers"
        trace.marker.size = 5
        trace.on_selection(self.select)
        trace.on_click(self.pick_cells)

    
def invalidArgs(func, argdict):
    import inspect
    args = inspect.getfullargspec(func).args
    return set(argdict) - set(args)
    



from setuptools import setup, find_packages

setup(
    name="sam-algorithm",
    version="0.8.8",
    description="The Self-Assembling-Manifold algorithm",
    long_description="The Self-Assembling-Manifold algorithm for analyzing single-cell RNA sequencing data.",
    long_description_content_type="text/markdown",
    url="https://github.com/atarashansky/self-assembling-manifold",
    author="Alexander J. Tarashansky",
    author_email="tarashan@stanford.edu",
    keywords="scrnaseq analysis manifold reconstruction",
    python_requires=">=3.6",
    # py_modules=["SAM", "utilities", "SAMGUI"],
    install_requires=[
        "numpy>=1.19.0",
        "scipy>=1.3.1",
        "pandas>1.0.0",
        "scikit-learn>=0.23.1",
        "packaging>=0.20.0",
        "numba>=0.50.1",
        "umap-learn>=0.4.6",
        "dill",
        "anndata>=0.7.4",
        "h5py<=2.10.0"
    ],
    extras_require={
        "louvain": ["louvain", "cython", "python-igraph"],
        "leiden": ["leidenalg", "cython", "python-igraph"],
        "hdbscan": ["hdbscan"],
        "plot": ["ipythonwidgets", "jupyter", "plotly==4.0.0", "matplotlib"],
        "scanpy": ["scanpy"],
    },
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
author:     Alexander Tarashansky
date:       10/01/2018
content:    Tests for SAM.
"""
# Modules
from samalg import SAM


# Script
if __name__ == "__main__":

    sam = SAM()
    sam.load_data("example_data/darmanis_data.csv.gz")
    sam.load_obs_annotations("example_data/darmanis_ann.csv")
    sam.preprocess_data()
    sam.run(projection=None)
    sam.run_umap()
    sam.run_tsne()
    sam.kmeans_clustering(4)
    sam.identify_marker_genes_ratio()
    sam.identify_marker_genes_rf()
    umap_obj = sam.umap_obj  # checking to see if umap_obj is stored in sam object



import pandas as pd
import numpy as np
import scipy as sp
import os
import errno
from sklearn.decomposition import PCA
import umap.distances as dist
from sklearn.utils.extmath import svd_flip
from sklearn.utils import check_array, check_random_state
from scipy import sparse
import sklearn.utils.sparsefuncs as sf
from umap.umap_ import nearest_neighbors
__version__ = "0.8.7"


def find_corr_genes(sam, input_gene):
    """Rank genes by their spatially averaged expression pattern correlations to
    a desired gene.

    Parameters
    ----------

    sam - SAM
        The analyzed SAM object

    input_gene - string
        The gene ID with respect to which correlations will be computed.

    Returns
    -------
    A ranked list of gene IDs based on correlation to the input gene.
    """
    all_gene_names = np.array(list(sam.adata.var_names))

    D_avg = sam.adata.layers["X_knn_avg"]

    input_gene = np.where(all_gene_names == input_gene)[0]

    if input_gene.size == 0:
        print(
            "Gene note found in the filtered dataset. Note "
            "that genes are case sensitive."
        )
        return

    pw_corr = generate_correlation_map(D_avg.T.A, D_avg[:, input_gene].T.A)
    return all_gene_names[np.argsort(-pw_corr.flatten())]

def _pca_with_sparse(X, npcs, solver='arpack', mu=None, seed=0):
    random_state = check_random_state(seed)
    np.random.set_state(random_state.get_state())
    random_init = np.random.rand(np.min(X.shape))
    X = check_array(X, accept_sparse=['csr', 'csc'])

    if mu is None:
        mu = X.mean(0).A.flatten()[None, :]
    mdot = mu.dot
    mmat = mdot
    mhdot = mu.T.dot
    mhmat = mu.T.dot
    Xdot = X.dot
    Xmat = Xdot
    XHdot = X.T.conj().dot
    XHmat = XHdot
    ones = np.ones(X.shape[0])[None, :].dot

    def matvec(x):
        return Xdot(x) - mdot(x)

    def matmat(x):
        return Xmat(x) - mmat(x)

    def rmatvec(x):
        return XHdot(x) - mhdot(ones(x))

    def rmatmat(x):
        return XHmat(x) - mhmat(ones(x))

    XL = sp.sparse.linalg.LinearOperator(
        matvec=matvec,
        dtype=X.dtype,
        matmat=matmat,
        shape=X.shape,
        rmatvec=rmatvec,
        rmatmat=rmatmat,
    )

    u, s, v = sp.sparse.linalg.svds(XL, solver=solver, k=npcs, v0=random_init)
    u, v = svd_flip(u, v)
    idx = np.argsort(-s)
    v = v[idx, :]

    X_pca = (u * s)[:, idx]
    ev = s[idx] ** 2 / (X.shape[0] - 1)

    total_var = sf.mean_variance_axis(X, axis=0)[1].sum()
    ev_ratio = ev / total_var

    output = {
        'X_pca': X_pca,
        'variance': ev,
        'variance_ratio': ev_ratio,
        'components': v,
    }
    return output



def nearest_neighbors_wrapper(X,n_neighbors=15,metric='correlation',metric_kwds={},angular=True,random_state=0):
    random_state=np.random.RandomState(random_state)
    return nearest_neighbors(X,n_neighbors,metric,metric_kwds,angular,random_state)[:2]

def knndist(nnma):
    x, y = nnma.nonzero()
    data = nnma.data
    knn = y.reshape((nnma.shape[0], nnma[0, :].data.size))
    val = data.reshape(knn.shape)
    return knn, val


def save_figures(filename, fig_IDs=None, **kwargs):
    """
    Save figures.

    Parameters
    ----------
    filename - str
        Name of output file

    fig_IDs - int, numpy.array, list, optional, default None
        A list of open figure IDs or a figure ID that will be saved to a
        pdf/png file respectively.

    **kwargs -
        Extra keyword arguments passed into 'matplotlib.pyplot.savefig'.

    """
    import matplotlib.pyplot as plt

    if fig_IDs is not None:
        if type(fig_IDs) is list:
            savetype = "pdf"
        else:
            savetype = "png"
    else:
        savetype = "pdf"

    if savetype == "pdf":
        from matplotlib.backends.backend_pdf import PdfPages

        if len(filename.split(".")) == 1:
            filename = filename + ".pdf"
        else:
            filename = ".".join(filename.split(".")[:-1]) + ".pdf"

        pdf = PdfPages(filename)

        if fig_IDs is None:
            figs = [plt.figure(n) for n in plt.get_fignums()]
        else:
            figs = [plt.figure(n) for n in fig_IDs]

        for fig in figs:
            fig.savefig(pdf, format="pdf", **kwargs)
        pdf.close()
    elif savetype == "png":
        plt.figure(fig_IDs).savefig(filename, **kwargs)


def weighted_PCA(mat, do_weight=True, npcs=None, solver="auto",seed = 0):
    # mat = (mat - np.mean(mat, axis=0))
    if do_weight:
        if min(mat.shape) >= 10000 and npcs is None:
            print(
                "More than 10,000 cells. Running with 'npcs' set to < 1000 is"
                " recommended."
            )

        if npcs is None:
            ncom = min(mat.shape)
        else:
            ncom = min((min(mat.shape), npcs))
            
        pca = PCA(svd_solver=solver, n_components=ncom,random_state=check_random_state(seed))
        reduced = pca.fit_transform(mat)
        scaled_eigenvalues = pca.explained_variance_
        scaled_eigenvalues = scaled_eigenvalues / scaled_eigenvalues.max()
        reduced_weighted = reduced * scaled_eigenvalues[None, :] ** 0.5
    else:
        pca = PCA(n_components=npcs, svd_solver=solver,random_state=check_random_state(seed))
        reduced = pca.fit_transform(mat)
        if reduced.shape[1] == 1:
            pca = PCA(n_components=2, svd_solver=solver,random_state=check_random_state(seed))
            reduced = pca.fit_transform(mat)
        reduced_weighted = reduced

    return reduced_weighted, pca


def transform_wPCA(mat, pca):
    mat = mat - pca.mean_
    reduced = mat.dot(pca.components_.T)
    v = pca.explained_variance_  # .var(0)
    scaled_eigenvalues = v / v.max()
    reduced_weighted = np.array(reduced) * scaled_eigenvalues[None, :] ** 0.5
    return reduced_weighted

def search_string(vec, s, case_sensitive=False, invert=False):
    vec = np.array(vec)


    if isinstance(s,list):
        S = s
    else:
        S = [s]

    V=[]; M=[]
    for s in S:
        m = []
        if not case_sensitive:
            s = s.lower()
        for i in range(len(vec)):
            if case_sensitive:
                st = vec[i]
            else:
                st = vec[i].lower()
            b = st.find(s)
            if not invert and b != -1 or invert and b == -1:
                m.append(i)
        if len(m) > 0:
            V.append(vec[np.array(m)]); M.append(np.array(m))
    if len(V)>0:
        i = len(V)
        if not invert:
            V = np.concatenate(V); M = np.concatenate(M);
            if i > 1:
                ix = np.sort(np.unique(M,return_index=True)[1])
                V=V[ix]; M=M[ix];
        else:
            for i in range(len(V)):
                V[i]=list(set(V[i]).intersection(*V))
            V = vec[np.in1d(vec,np.unique(np.concatenate(V)))]
            M = np.array([np.where(vec==x)[0][0] for x in V])
        return V,M
    else:
        return -1,-1


def distance_matrix_error(dist1, dist2):
    s = 0
    for k in range(dist1.shape[0]):
        s += np.corrcoef(dist1[k, :], dist2[k, :])[0, 1]
    return 1 - s / dist1.shape[0]


def generate_euclidean_map(A, B):
    a = (A ** 2).sum(1).flatten()
    b = (B ** 2).sum(1).flatten()
    x = a[:, None] + b[None, :] - 2 * np.dot(A, B.T)
    x[x < 0] = 0
    return np.sqrt(x)


def generate_correlation_map(x, y):
    mu_x = x.mean(1)
    mu_y = y.mean(1)
    n = x.shape[1]
    if n != y.shape[1]:
        raise ValueError("x and y must " + "have the same number of timepoints.")
    s_x = x.std(1, ddof=n - 1)
    s_y = y.std(1, ddof=n - 1)
    s_x[s_x == 0] = 1
    s_y[s_y == 0] = 1
    cov = np.dot(x, y.T) - n * np.dot(mu_x[:, None], mu_y[None, :])
    return cov / np.dot(s_x[:, None], s_y[None, :])


def extract_annotation(cn, x, c="_"):
    m = []
    if x is not None:
        for i in range(cn.size):
            f = cn[i].split(c)
            x = min(len(f) - 1, x)
            m.append(f[x])
        return np.array(m)
    else:
        ms = []
        ls = []
        for i in range(cn.size):
            f = cn[i].split(c)
            m = []
            for x in range(len(f)):
                m.append(f[x])
            ms.append(m)
            ls.append(len(m))
        ml = max(ls)
        for i in range(len(ms)):
            ms[i].extend([""] * (ml - len(ms[i])))
            if ml - len(ms[i]) > 0:
                ms[i] = np.concatenate(ms[i])
        ms = np.vstack(ms)
        MS = []
        for i in range(ms.shape[1]):
            MS.append(ms[:, i])
        return MS


def isolate(dt, x1, x2, y1, y2):
    return np.where(
        np.logical_and(
            np.logical_and(dt[:, 0] > x1, dt[:, 0] < x2),
            np.logical_and(dt[:, 1] > y1, dt[:, 1] < y2),
        )
    )[0]


def to_lower(y):
    x = y.copy().flatten()
    for i in range(x.size):
        x[i] = x[i].lower()
    return x


def to_upper(y):
    x = y.copy().flatten()
    for i in range(x.size):
        x[i] = x[i].upper()
    return x


def create_folder(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def convert_annotations(A):
    x = np.unique(A)
    y = np.zeros(A.size)
    z = 0
    for i in x:
        y[A == i] = z
        z += 1

    return y.astype("int")


def nearest_neighbors_hnsw(x,ef=200,M=48,n_neighbors = 100):
    import hnswlib
    labels = np.arange(x.shape[0])
    p = hnswlib.Index(space = 'cosine', dim = x.shape[1])
    p.init_index(max_elements = x.shape[0], ef_construction = ef, M = M)
    p.add_items(x, labels)
    p.set_ef(ef)
    idx, dist = p.knn_query(x, k = n_neighbors)
    return idx,dist


def calc_nnm(g_weighted, k, distance=None):
    if g_weighted.shape[0] > 0:
        if distance == 'cosine':
            nnm, dists = nearest_neighbors_hnsw(g_weighted, n_neighbors=k)
        else:
            nnm, dists = nearest_neighbors_wrapper(g_weighted, n_neighbors=k, metric=distance)
        EDM = gen_sparse_knn(nnm, dists)
        EDM = EDM.tocsr()        
    return EDM


def compute_distances(A, dm):
    if dm == "euclidean":
        m = np.dot(A, A.T)
        h = np.diag(m)
        x = h[:, None] + h[None, :] - 2 * m
        x[x < 0] = 0
        dist = np.sqrt(x)
    elif dm == "correlation":
        dist = 1 - np.corrcoef(A)
    else:
        dist = sp.spatial.distance.squareform(sp.spatial.distance.pdist(A, metric=dm))
    return dist


def dist_to_nn(d, K):  # , offset = 0):
    E = d.copy()
    np.fill_diagonal(E, -1)
    M = np.max(E) * 2
    x = np.argsort(E, axis=1)[:, :K]  # offset:K+offset]
    E[
        np.tile(
            np.arange(E.shape[0]).reshape(E.shape[0], -1), (1, x.shape[1])
        ).flatten(),
        x.flatten(),
    ] = M

    E[E < M] = 0
    E[E > 0] = 1
    return E  # ,x


def to_sparse_knn(D1, k):
    for i in range(D1.shape[0]):
        x = D1.data[D1.indptr[i] : D1.indptr[i + 1]]
        idx = np.argsort(x)
        if idx.size > k:
            x[idx[:-k]] = 0
        D1.data[D1.indptr[i] : D1.indptr[i + 1]] = x
    D1.eliminate_zeros()
    return D1


def gen_sparse_knn(knni, knnd, shape=None):
    if shape is None:
        shape = (knni.shape[0], knni.shape[0])

    D1 = sp.sparse.lil_matrix(shape)

    D1[
        np.tile(np.arange(knni.shape[0])[:, None], (1, knni.shape[1])).flatten().astype('int32'),
        knni.flatten().astype('int32'),
    ] = knnd.flatten()
    D1 = D1.tocsr()
    return D1
